#!/usr/bin/env python3
#
# Test for verifying whether once a long, complex sequence has been started, it can run to completion and/or be successfully aborted
#
# Test procedure for abortion:
# - start marcos_server on hardware
# - run this file - it just slowly increments the LEDs and does a loopback for a long time
# - halt marcos_server explicitly on hardware; this file will crash due to the server going down
# - restart marcos_server
# - run this file again, and verify that the LED pattern starts again from 0
#
# If you wish to see what happens for different delays, just alter the arguments to long_loopback() at the end of this file.

import numpy as np
import matplotlib.pyplot as plt
import experiment as ex
from local_config import grad_board

import pdb
st = pdb.set_trace

# In each TR, a wait of rf_start_delay occurs, the RF turns on, it
# turns off after rf_length. Each TR is by default 1 second long. The RX
# turns on rx_pad microseconds before and turns off rx_pad
# microseconds after the RF pulse.

lo_freq = 5 # MHz
rx_period = 5 # us
rx_pad = 20 # us
rf_start_delay = 100 # us
rf_amp = 0.4
rf_length = 200 # us

def long_loopback(rf_interval=1000000, trs=20):

    expt = ex.Experiment(lo_freq=lo_freq, rx_t=rx_period, halt_and_reset=True)

    for k in range(trs):
        rf_t = rf_start_delay + k*rf_interval + np.array([0, rf_length])
        rx_t = rf_start_delay + k*rf_interval - rx_pad + np.array([0, rf_length + 2*rx_pad])
        expt.add_flodict({
            'tx1': ( rf_t, np.array([rf_amp, 0]) ),
            'rx1_en': ( rx_t, np.array([1, 0]) )
            # 'leds': ( np.array([k*rf_interval]), np.array(k) )
            })

    rxd, msgs = expt.run()
    expt.close_server(only_if_sim=True)

    expt._s.close() # close socket on client

if __name__ == "__main__":
    long_loopback(1000000, 255) #



#!/usr/bin/env python3
#
# loopback test using ocra-pulseq
#

import numpy as np
import matplotlib.pyplot as plt
import pdb, sys
import experiment as ex
from pulseq_assembler import PSAssembler
st = pdb.set_trace

if __name__ == "__main__":
    lo_freq = 2.1 # MHz
    tx_t = 1.001 # us
    clk_t = 0.007 # ocra system clk for tx and gradients -- rx system clk depends on RP/STEMlab model
    num_grad_channels = 3
    grad_interval = 10.003 # us between [num_grad_channels] channel updates

    do_single_test = True
    do_jitter_test = False

    if len(sys.argv) > 1 and "jitter" in sys.argv:
        do_jitter_test = True
        do_single_test = False
    
    if False: # VN: this stuff belongs in a more sophisticated test, not test_loopback. TODO: refactor/remove
        gamma = 42570000 # Hz/T

        # value for tabletopMRI  gradient coil
        B_per_m_per_current = 0.02 # T/m/A, approximate value for tabletop gradient coil

        # values for gpa fhdo
        gpa_current_per_volt = 3.75 # A/V, theoretical value for gpa fhdo without 2.5 V offset!
        max_dac_voltage = 2.5

        # values for ocra1
        # max_dac_voltage = 5
        # gpa_current_per_volt = 3.75 # A/V, fill in value of gradient power amplifier here!

        max_Hz_per_m = max_dac_voltage * gpa_current_per_volt * B_per_m_per_current * gamma	
        # grad_max = max_Hz_per_m # factor used to normalize gradient amplitude, should be max value of the gpa used!	

    # factor used to normalise RF amplitude, should be max value in system used!
    # (I.e. if unknown, lower values will create larger RF signals, since ocra-pulseq normalises its outputs to this value)
    rf_amp_max = 16

    # as above - for GPA boards, should represent full-scale voltage (or current, if that's being measured)
    # ocra-pulseq normalises its outputs to this value
    grad_max = 10

    ps = PSAssembler(rf_center=lo_freq*1e6,
                     rf_amp_max=rf_amp_max,
                     grad_max=grad_max,
                     clk_t=clk_t,
                     tx_t=tx_t,
                     grad_t=grad_interval,
                     grad_pad=2,
                     addresses_per_grad_sample=3)
    tx_arr, grad_arr, cb, params = ps.assemble('../ocra-pulseq/test_files/test_loopback.seq', byte_format=False)
    
    exp = ex.Experiment(samples=params['readout_number'], # TODO: get information from PSAssembler
                        lo_freq=lo_freq,
                        tx_t=tx_t,
                        rx_t=params['rx_t'],
                        grad_channels=num_grad_channels,
                        grad_t=grad_interval/num_grad_channels,
                        assert_errors=False,
                        print_infos=True)
    
    exp.define_instructions(cb)

    if False: # test sine for grad/tx
        x = np.linspace(0,2*np.pi, 100)
        ramp_sine = np.sin(2*x)
    
    exp.add_tx(tx_arr)
    exp.add_grad(grad_arr)

    if do_single_test:
        exp.run()
        
    if do_jitter_test:
        data = []
        trials = 1000
        for k in range(trials):
            d, s = exp.run()
            # TODO: retake when warnings occur due to timeouts etc
            data.append( d ) # Comment out this line to avoid running on the hardware
        
        taxis = np.arange(params['readout_number'])*params['rx_t']
        plt.figure(figsize=(10,9))

        good_data = []
        bad_data = []

        for d in data:
            if np.abs(d[-1]) == 0.0:
                bad_data.append(d)
            else:
                good_data.append(d)

        lgd = len(good_data)
        lbd = len(bad_data)
                
        plt.subplot(2,1,1)
        for d in good_data:
            plt.plot(taxis, d.real )
        plt.ylabel('loopback rx amplitude')
        plt.title('passing loopback data ({:d}/{:d}, {:.2f}%)'.format(lgd, lgd+lbd, 100*lgd/(lgd+lbd)))
        plt.grid(True)

        plt.subplot(2,1,2)
        for d in bad_data:
            plt.plot(taxis, d.real )
        plt.ylabel('loopback rx amplitude')
        plt.title('failing loopback data ({:d}/{:d}, {:.2f}%)'.format(lbd, lgd+lbd, 100*lbd/(lgd+lbd)))
        plt.grid(True)        
        
        plt.show()



import os
import sys
import csv
from marge.configs import hw_config as hw
from PyQt5.QtWidgets import (
    QApplication, QWidget, QVBoxLayout, QLineEdit, QPushButton,
    QHBoxLayout, QLabel, QGridLayout
)


class ConsoleWidget(QWidget):
    def __init__(self):
        super().__init__()

        # Main layout
        self.main_layout = QVBoxLayout()
        self.dynamic_container = QVBoxLayout()  # Store reference
        self.layout = QGridLayout()

        # Parameters to save inputs
        self.labels = []
        self.values = []
        self.tips = []

        # Add inputs
        self.add_input(label="Red Pitaya model", value="rp-122", tip="Model of the Red Pitaya board")
        self.add_input(label="Maximum input voltage (mV)", value="225",
                       tip="Maximum voltage that can be input to the system")
        self.add_input(label="Gradient board model", value="gpa-fhdo", tip="Model of the gradient controller board")
        self.add_input(label="Clock frequency (MHz)", value="122.88", tip="Clock frequency of the system")
        self.add_input(label="ADC factor (mV/unit)", value="13.788", tip="ADC conversion factor from unit to mV")
        self.add_input(label="CIC delay points", value="3", tip="Number of delay points in the CIC filter")

        # Dictionary to store references to input fields
        self.input_boxes = {}

        # Create blocks iteratively
        for row, (label, value) in enumerate(zip(self.labels, self.values)):
            label_widget = QLabel(label)
            input_box = QLineEdit(value)
            input_box.setStatusTip(self.tips[row])
            self.input_boxes[label] = input_box

            self.layout.addWidget(label_widget, row, 0)  # Label in column 0
            self.layout.addWidget(input_box, row, 1)  # Input box in column 1

        # Input field for RP ip address
        self.text_box = QLineEdit(self)
        self.text_box.setPlaceholderText("Red Pitaya ip address")

        # Buttons
        self.add_button = QPushButton('Add', self)
        self.add_button.clicked.connect(self.add_rp)

        self.save_button = QPushButton('Save', self)
        self.save_button.clicked.connect(self.save_rp_entries)

        # self.layout.addLayout(self.dynamic_container)

        layout = QHBoxLayout()
        layout.addWidget(self.text_box)
        layout.addWidget(self.add_button)
        layout.addWidget(self.save_button)
        self.main_layout.addLayout(self.layout)
        self.main_layout.addLayout(self.dynamic_container)
        self.main_layout.addLayout(layout)
        self.main_layout.addStretch()

        self.setLayout(self.main_layout)
        self.setWindowTitle('Red Pitaya Entry')
        self.resize(400, 200)

        # Counter for IDs
        self.rp_counter = 1

        # Store added RPs in a list
        self.added_rps = []

        # Load saved RP entries
        self.load_rp_entries()

        # Update hardware
        self.update_hw_config_rp()

    def add_input(self, label="", value="", tip=""):
        self.labels.append(label)
        self.values.append(value)
        self.tips.append(tip)

    def update_hw_config_rp(self):
        hw.rp_ip_list = []
        hw.rp_port = []
        for _, ip in self.added_rps:
            hw.rp_ip_list.append(ip)
            hw.rp_port.append(11111)
        hw.rp_version = self.input_boxes["Red Pitaya model"].text()
        hw.rp_max_input_voltage = float(self.input_boxes["Maximum input voltage (mV)"].text())
        hw.grad_board = self.input_boxes["Gradient board model"].text()
        hw.fpga_clk_freq_MHz = float(self.input_boxes["Clock frequency (MHz)"].text())
        hw.adcFactor = float(self.input_boxes["ADC factor (mV/unit)"].text())
        hw.cic_delay_points = int(self.input_boxes["CIC delay points"].text())

    def add_rp(self):
        text = self.text_box.text().strip()
        if text:
            row_layout = QHBoxLayout()
            identifier_label = QLabel(f"RP-{self.rp_counter}")  # Unique identifier
            text_label = QLabel(text)  # User input

            # Add Delete button for this RP entry
            delete_button = QPushButton('Delete')
            delete_button.clicked.connect(lambda: self.delete_rp(row_layout))

            row_layout.addWidget(identifier_label)
            row_layout.addWidget(text_label)
            row_layout.addWidget(delete_button)

            self.dynamic_container.addLayout(row_layout)  # Add to container

            # Add the RP to the list for saving/deleting purposes
            self.added_rps.append((self.rp_counter, text))

            self.rp_counter += 1  # Increment counter
            self.text_box.clear()  # Clear input field

    def delete_rp(self, row_layout):
        # Remove the RP entry from the container and list
        for i, (rp_id, _) in enumerate(self.added_rps):
            if row_layout.itemAt(0).widget().text() == f"RP-{rp_id}":
                self.added_rps.pop(i)
                break

        # Remove layout (delete UI elements)
        for i in reversed(range(row_layout.count())):
            widget = row_layout.itemAt(i).widget()
            if widget is not None:
                widget.deleteLater()

        self.dynamic_container.update()

    def save_rp_entries(self):
        file_name = "configs/hw_redpitayas.csv"
        with open(file_name, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(["ID", "Value"])
            for label, input_box in self.input_boxes.items():
                writer.writerow([label, input_box.text()])  # Write each pair
            for rp_id, ip in self.added_rps:
                writer.writerow([f"RP-{rp_id}", ip])
        self.update_hw_config_rp()
        print(f"Data saved for red pitayas")

    def load_rp_entries(self):
        file_name = "configs/hw_redpitayas.csv"
        if os.path.exists(file_name):
            with open(file_name, 'r') as csvfile:
                reader = csv.reader(csvfile)
                next(reader)  # Skip header row
                for row in reader:
                    label, value = row
                    if label in self.input_boxes:
                        self.input_boxes[label].setText(value)
                    else:
                        self.added_rps.append((label.split('-')[1], value))
                        self.add_rp_from_data(label.split('-')[1], value)
        else:
            print("No hardware configuration loaded for red pitayas.")

    def add_rp_from_data(self, rp_id, ip):
        row_layout = QHBoxLayout()
        identifier_label = QLabel(f"RP-{rp_id}")
        text_label = QLabel(ip)

        # Add Delete button for this RP entry
        delete_button = QPushButton('Delete')
        delete_button.clicked.connect(lambda: self.delete_rp(row_layout))

        row_layout.addWidget(identifier_label)
        row_layout.addWidget(text_label)
        row_layout.addWidget(delete_button)

        self.dynamic_container.addLayout(row_layout)  # Add to container


if __name__ == '__main__':
    # Run the application
    app = QApplication(sys.argv)
    widget = ConsoleWidget()
    widget.show()
    sys.exit(app.exec())


"""
Created on Thu June 2 2022
@author: J.M. Algarín, MRILab, i3M, CSIC, Valencia
@email: josalggui@i3m.upv.es
@Summary: All sequences on the GUI must be here
"""

import inspect
import os
import importlib

"""
Definition of default sequences
"""

# Note for the users: Now the sequences are added automatically to the defaultsequences dictionary.
# To do that, the user should include the parameter 'toMaRGE' as True in the sequence using:
# self.addParameter(string='toMaRGE', val=True)
# This file should not be modified anymore.

def instantiate_sequences():
    # Get the absolute path to this folder (marge/seq)
    folder = os.path.dirname(__file__)

    # Dictionary to store class sequences
    defaultsequences = {}

    # List all .py files in the folder, except __init__.py
    py_files = [f for f in os.listdir(folder) if f.endswith('.py') and f != '__init__.py']

    # Populate defaultsequences
    for file in py_files:
        # Remove the .py extension to get the module name
        module_name = file[:-3]

        try:
            # Dynamically import the module using full path
            module = importlib.import_module(f"marge.seq.{module_name}")

            # Find all classes in the module
            classes = inspect.getmembers(module, inspect.isclass)

            # Add to defaultsequences only if toMaRGE is True
            for class_name, class_ in classes:
                try:
                    if class_().mapVals['toMaRGE']:
                        defaultsequences[class_().mapVals['seqName']] = class_()
                        print(f"{class_().mapVals['seqName']} added to MaRGE")
                except:
                    pass
        except Exception as e:
            print(f"Error importing module {module_name}: {e}")

    return defaultsequences

defaultsequences = instantiate_sequences()

# Note for the users: Now the sequences are added automatically to the defaultsequences dictionary.
# To do that, the user should include the parameter 'toMaRGE' as True in the sequence using:
# self.addParameter(string='toMaRGE', val=True)
# This file should not be modified anymore.



from PyQt5.QtWidgets import QPushButton, QLabel, QLineEdit, QHBoxLayout, QVBoxLayout, QWidget


class VisualisationTabWidget(QWidget):
    """
    VisualisationTabWidget class for displaying a tab widget for visualization settings.

    Inherits from QTabWidget provided by PyQt5 to display a tab widget for visualization settings.

    Attributes:
        main: The parent widget.
    """

    def __init__(self, parent, *args, **kwargs):
        """
        Initialize the VisualisationTabWidget.

        Args:
            parent: The parent widget.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(VisualisationTabWidget, self).__init__(*args, **kwargs)
        self.main = parent

        # Button
        self.visualisation_button = QPushButton('Show slices')

        # Labels
        self.range_label = QLabel('Slices')
        self.column_label = QLabel('Rows and columns')

        # Text Fields
        self.range_text_field = QLineEdit()
        self.range_text_field.setPlaceholderText('First, Last')

        self.column_text_field = QLineEdit()
        self.column_text_field.setPlaceholderText('Rows, Columns')

        # Layouts
        self.number_layout = QHBoxLayout()
        self.number_layout.addWidget(self.range_label)
        self.number_layout.addWidget(self.range_text_field)

        self.column_layout = QHBoxLayout()
        self.column_layout.addWidget(self.column_label)
        self.column_layout.addWidget(self.column_text_field)

        self.visualisation_layout = QVBoxLayout()
        self.visualisation_layout.addLayout(self.number_layout)
        self.visualisation_layout.addLayout(self.column_layout)
        self.visualisation_layout.addWidget(self.visualisation_button)
        self.visualisation_layout.addStretch()

        self.setLayout(self.visualisation_layout)



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import numpy as np

from marge.seq.sequences import defaultsequences
from marge.widgets.widget_plot3d import Plot3DWidget
import marge.configs.hw_config as hw


class Plot3DController(Plot3DWidget):
    """
    Controller class for a 3D plot widget.
    """
    def __init__(self, data=np.array([]), x_label='', y_label='', title='', *args, **kwargs):
        """
        Initialize the Plot3DController.

        Args:
            data (ndarray): The 3D data array for the plot.
            x_label (str): The label for the x-axis.
            y_label (str): The label for the y-axis.
            title (str): The title of the plot.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        super(Plot3DController, self).__init__(*args, **kwargs)
        self.seq_name = None
        self.data = data
        self.x_label = x_label
        self.y_label = y_label
        self.title = title
        self.img_resolution = None

        # Set plot_item properties
        self.plot_item.setLabel(axis='left', text=y_label)
        self.plot_item.setLabel(axis='bottom', text=x_label)
        self.plot_item.setTitle(title=title)

        # hide FOV button if not image
        self.ui.menuBtn.hide()
        if title == "Sagittal" or title == "Coronal" or title == "Transversal":
            self.ui.menuBtn.show()

        # Change the Norm button to FOV button
        self.ui.menuBtn.setText("FOV")
        self.ui.menuBtn.setCheckable(True)

        # Create ROI to get FOV
        self.roiFOV.setSize(np.size(data, 1), np.size(data, 2), update=False)
        self.roiFOV.addScaleHandle([1, 1], [0.5, 0.5])
        self.roiFOV.addRotateHandle([0, 0], [0.5, 0.5])
        self.roiFOV.setZValue(20)
        self.roiFOV.setPen('y')
        self.roiFOV.hide()
        self.roiFOV.sigRegionChangeFinished.connect(self.roiFOVChanged)

        # Modify ROI to get SNR
        self.roi.setPen('g')

        # Add image
        self.setImage(data)

    def menuClicked(self):
        """
        Handle the FOV button click event.

        This method is called when the FOV button is clicked.

        """
        # Now the menu button is the FOV button

        # Provisionally stop signals from roi changes
        for widget in self.main.figures_layout.findChildren(Plot3DController):
            widget.roiFOV.blockSignals(True)

        if self.ui.menuBtn.isChecked():
            # Get the corresponding axes from the image
            x_axis = 0
            y_axis = 1
            z_axis = 2
            d = [1, 1]
            a = 1
            if self.title == "Sagittal":
                a = -1
                d = [1, 1] #[PA, IS]
                x_axis = 1
                y_axis = 0
                z_axis = 2
            elif self.title == "Coronal":
                a = -1
                d = [1, 1] #[RL, IS]
                x_axis = 2
                y_axis = 0
                z_axis = 1
            elif self.title == "Transversal":
                a = 1
                d = [1, 1] #[RL, PA]
                x_axis = 2
                y_axis = 1
                z_axis = 0

            # Get image fov and resolution
            current_output = self.main.history_list.current_output
            self.seq_name = self.main.history_list.inputs[current_output][1][0]
            img_fov = self.main.history_list.fovs[current_output][-1]
            if self.seq_name in defaultsequences.keys():
                roi_fov = np.array(defaultsequences[self.seq_name].mapVals['fov']) * 1e-2  # m
                roi_pos = np.array(defaultsequences[self.seq_name].mapVals['dfov']) * 1e-3  # m
            else:
                roi_fov = img_fov
                roi_pos = [0.0, 0.0, 0.0]

            img_fov_px = np.array(np.shape(self.getProcessedImage()))[1::]
            img_fov_ru = [img_fov[x_axis], img_fov[y_axis]]
            self.img_resolution = np.array(img_fov_ru) / img_fov_px
            roi_fov_ru = np.array([roi_fov[x_axis], roi_fov[y_axis]])
            roi_pos_ru = np.array([d[0] * roi_pos[x_axis], d[1] * roi_pos[y_axis]])
            roi_fov_px = np.array(roi_fov_ru) / self.img_resolution
            roi_pos_px = np.array(roi_pos_ru) / self.img_resolution
            roi_pos_px = roi_pos_px + img_fov_px / 2 - roi_fov_px / 2

            # Set roi size and angle
            self.roiFOV.setPos(roi_pos_px, update=False)
            self.roiFOV.setSize(roi_fov_px, update=False)
            self.roiFOV.setAngle(0.0, update=False)
            self.roiFOV.stateChanged()

            # Show the roi
            self.roiFOV.show()
        else:
            self.roiFOV.hide()

        # Reset signals from roi changes
        for widget in self.main.figures_layout.findChildren(Plot3DController):
            widget.roiFOV.blockSignals(False)

    def roiFOVChanged(self):
        """
        Update sequence parameters and propagate changes to other widgets based on the region of interest (ROI) field
        of view (FOV).

        This method calculates the necessary parameters for the ROI FOV, updates the sequence parameters, and ensures
        that the changes are reflected in other widgets.

        """
        # Get the corresponding axes from the image
        x_axis = 0
        y_axis = 1
        z_axis = 2
        d = [1, 1]
        a = 1
        if self.title == "Sagittal":
            a = -1
            d = [1, 1]
            x_axis = 1
            y_axis = 0
            z_axis = 2
        elif self.title == "Coronal":
            a = 1
            d = [1, 1]
            x_axis = 2
            y_axis = 0
            z_axis = 1
        elif self.title == "Transversal":
            a = 1
            d = [1, 1]
            x_axis = 2
            y_axis = 1
            z_axis = 0

        # Get roi properties in pixel units
        ima_fov_px = np.array(np.shape(self.getProcessedImage()))[1::]
        roi_fov_px = self.roiFOV.size()
        roi_pos_px = self.roiFOV.pos()
        roi_angle = self.roiFOV.angle()

        # ROI center in pixel units
        x0 = roi_pos_px[0]  # Upper left corner
        y0 = roi_pos_px[1]  # Upper left corner
        x1 = x0 + roi_fov_px[0] * np.cos(- roi_angle * np.pi / 180)  # Upper right corner
        y1 = y0 - roi_fov_px[0] * np.sin(- roi_angle * np.pi / 180)  # Upper right corner
        x2 = x1 + roi_fov_px[1] * np.sin(- roi_angle * np.pi / 180)  # Bottom right corner
        y2 = y1 + roi_fov_px[1] * np.cos(- roi_angle * np.pi / 180)  # Bottom right corner
        x3 = (x0 + x2) / 2  # Center
        y3 = (y0 + y2) / 2  # Center

        # ROI center in real units
        x0_ru = (x3 - ima_fov_px[0] / 2) * self.img_resolution[0]
        y0_ru = (y3 - ima_fov_px[1] / 2) * self.img_resolution[1]

        # Set fov properties in true units
        fov_roi = hw.fov.copy()
        fov_roi[x_axis] = np.round(roi_fov_px[0] * self.img_resolution[0] * 1e2, decimals=1)  # cm
        fov_roi[y_axis] = np.round(roi_fov_px[1] * self.img_resolution[1] * 1e2, decimals=1)  # cm
        dfov_roi = hw.dfov.copy()
        dfov_roi[x_axis] = d[0] * np.round(x0_ru * 1e3, decimals=1)  # mm
        dfov_roi[y_axis] = d[1] * np.round(y0_ru * 1e3, decimals=1)  # mm
        hw.fov[x_axis] = np.round(fov_roi[x_axis])
        hw.fov[y_axis] = np.round(fov_roi[y_axis])
        hw.dfov[x_axis] = np.round(dfov_roi[x_axis])
        hw.dfov[y_axis] = np.round(dfov_roi[y_axis])

        # Define rotation
        rotation = [0, 0, 0, 0]
        rotation[z_axis] = 1
        rotation[3] = a * roi_angle

        # Update sequence parameters
        for sequence in defaultsequences.values():
            if 'fov' in sequence.mapKeys:
                sequence.mapVals['fov'][x_axis] = np.round(fov_roi[x_axis], decimals=1)  # cm
                sequence.mapVals['fov'][y_axis] = np.round(fov_roi[y_axis], decimals=1)  # cm
            if 'dfov' in sequence.mapKeys:
                sequence.mapVals['dfov'][x_axis] = np.round(dfov_roi[x_axis], decimals=1)  # mm
                sequence.mapVals['dfov'][y_axis] = np.round(dfov_roi[y_axis], decimals=1)  # mm
            if 'angle' in sequence.mapKeys:
                sequence.mapVals['angle'] = np.round(rotation[3], decimals=2)  # degrees
            if 'rotationAxis' in sequence.mapKeys:
                sequence.mapVals['rotationAxis'] = rotation[0:3]

        # Loop over the widgets in the figure_layout to link the roiFOV
        for widget in self.main.figures_layout.findChildren(Plot3DController):
            if id(widget) == id(self):
                pass
            else:
                widget.menuClicked()
                if widget.title == 'Sagittal' or widget.title == 'Coronal' or widget.title == 'Transversal':
                    if roi_angle != 0:
                        print("Warning: it is not recommended to angle the figure with multiplot.")

        self.main.sequence_list.updateSequence()

    def roiChanged(self):
        """
        Update the plot and text items based on the selected region of interest (ROI).

        This method extracts the image data within the ROI, calculates necessary statistics, and updates the plot and
        text items accordingly.

        """
        # Extract image data from ROI
        if self.image is None:
            return

        image = self.getProcessedImage()

        # getArrayRegion axes should be (x, y) of data array for col-major,
        # (y, x) for row-major
        # can't just transpose input because ROI is axisOrder aware
        colmaj = self.imageItem.axisOrder == 'col-major'
        if colmaj:
            axes = (self.axes['x'], self.axes['y'])
        else:
            axes = (self.axes['y'], self.axes['x'])

        data, coords = self.roi.getArrayRegion(
            image.view(np.ndarray), img=self.imageItem, axes=axes,
            returnMappedCoords=True)

        if data is None:
            return

        # Convert extracted data into 1D plot data
        if self.axes['t'] is None:
            # Average across y-axis of ROI
            data = data.mean(axis=self.axes['y'])

            # Get average and std of current slice
            dataSlice = np.reshape(data, -1)  # Here I get the roi of current index
            self.dataAvg = dataSlice.mean()
            self.dataStd = dataSlice.std()
            self.dataSnr = self.dataAvg / self.dataStd

            # get coordinates along x axis of ROI mapped to range (0, roiwidth)
            if colmaj:
                coords = coords[:, :, 0] - coords[:, 0:1, 0]
            else:
                coords = coords[:, 0, :] - coords[:, 0, 0:1]
            xvals = (coords ** 2).sum(axis=0) ** 0.5
        else:
            # Get average and std of current slice
            (ind, time) = self.timeIndex(self.timeLine)
            dataSlice = np.reshape(data[ind, :, :], -1)  # Here I get the roi of current index
            self.dataAvg = dataSlice.mean()
            self.dataStd = dataSlice.std()
            self.dataSnr = self.dataAvg / self.dataStd

            # Average data within entire ROI for each frame
            mean = data.mean(axis=axes)
            std = data.std(axis=axes)
            data = mean / std
            xvals = self.tVals

        # Handle multi-channel data
        if data.ndim == 1:
            plots = [(xvals, data, 'w')]
        if data.ndim == 2:
            if data.shape[1] == 1:
                colors = 'w'
            else:
                colors = 'rgbw'
            plots = []
            for i in range(data.shape[1]):
                d = data[:, i]
                plots.append((xvals, d, colors[i]))

        # Update plot line(s)
        while len(plots) < len(self.roiCurves):
            c = self.roiCurves.pop()
            c.scene().removeItem(c)
        while len(plots) > len(self.roiCurves):
            self.roiCurves.append(self.ui.roiPlot.plot())
        for i in range(len(plots)):
            x, y, p = plots[i]
            self.roiCurves[i].setData(x, y, pen=p)

        # Update text_item
        self.text_item.setText("Mean = %0.1f \nstd = %0.1f \nsnr = %0.1f" % (self.dataAvg, self.dataStd, self.dataSnr))
        self.text_item.show()

    def roiClicked(self):
        """
        Handle the event when the ROI button is clicked.

        This method shows or hides the ROI plot based on the button state. It also updates the visibility of the ROI
        and the time line.

        """
        show_roi_plot = False
        if self.ui.roiBtn.isChecked():
            show_roi_plot = True
            self.roi.show()
            self.ui.roiPlot.setMouseEnabled(True, True)
            # self.ui.splitter.setSizes([int(self.height() * 0.6), int(self.height() * 0.4)])
            self.ui.splitter.handle(1).setEnabled(True)  # Allow to change the window size
            self.roiChanged()
            for c in self.roiCurves:
                c.hide()
            # self.ui.roiPlot.showAxis('left')
        else:
            self.roi.hide()
            self.ui.roiPlot.setMouseEnabled(False, False)
            for c in self.roiCurves:
                c.hide()
            self.ui.roiPlot.hideAxis('left')
            if hasattr(self, 'text_item'):
                self.text_item.hide()

        if self.hasTimeAxis():
            show_roi_plot = True
            mn = self.tVals.min()
            mx = self.tVals.max()
            self.ui.roiPlot.setXRange(mn, mx, padding=0.01)
            self.timeLine.show()
            self.timeLine.setBounds([mn, mx])
            if not self.ui.roiBtn.isChecked():
                # self.ui.splitter.setSizes([self.height() - 35, 35])
                self.ui.splitter.handle(1).setEnabled(False)
        else:
            self.timeLine.hide()

        self.ui.roiPlot.setVisible(show_roi_plot)

    def updateImage(self, autoHistogramRange=True):
        """
        Update and redraw the image on the screen.

        Args:
            autoHistogramRange (bool, optional): Flag indicating whether to automatically set the histogram range based on the level min and max. Defaults to True.

        """
        ## Redraw image on screen
        if self.image is None:
            return

        image = self.getProcessedImage()

        if autoHistogramRange:
            self.ui.histogram.setHistogramRange(self.levelMin, self.levelMax)

        # Transpose image into order expected by ImageItem
        if self.imageItem.axisOrder == 'col-major':
            axorder = ['t', 'x', 'y', 'c']
        else:
            axorder = ['t', 'y', 'x', 'c']
        axorder = [self.axes[ax] for ax in axorder if self.axes[ax] is not None]
        image = image.transpose(axorder)

        # Select time index
        if self.axes['t'] is not None:
            self.ui.roiPlot.show()
            image = image[self.currentIndex]
            if self.ui.roiBtn.isChecked():
                self.roiChanged()

        self.imageItem.updateImage(image)

    def hideAxis(self, axis):
        """
        Hide the specified axis in the plot.

        Args:
            axis (str): The axis to hide. Possible values are 'left', 'right', 'top', and 'bottom'.

        """
        self.plot_item.hideAxis(axis)

    def updateText(self, info):
        """
        Update the text item with the provided information.

        Args:
            info (str): The text to be displayed.

        """
        self.vbox.removeItem()
        self.text_item.setText(info)
        self.vbox.addItem(self.textitem)

    def setLabel(self, axis, text):
        """
        Set the label for the specified axis.

        Args:
            axis (str): The axis for which to set the label. Possible values are 'left', 'right', 'top', and 'bottom'.
            text (str): The label text.

        """
        self.plot_item.setLabel(axis=axis, text=text)

    def setTitle(self, title):
        """
        Set the title of the plot.

        Args:
            title (str): The title text.

        """
        self.plot_item.setTitle(title=title)

    def showHistogram(self, show=True):
        """
        Show or hide the histogram widget.

        Args:
            show (bool, optional): Flag indicating whether to show the histogram. Defaults to True.

        """
        hist = self.getHistogramWidget()
        hist.setVisible(show)



import csv
import os
import subprocess
import sys
from datetime import datetime

from PyQt5.QtWidgets import QWidget, QSizePolicy, QGridLayout, QComboBox, QInputDialog, QMessageBox, QApplication, \
    QLabel, QLineEdit

from marge.configs import hw_config, sys_config


class SessionWidget(QWidget):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Session Window")
        self.setSizePolicy(QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding))
        self.setMinimumWidth(400)

        # Add layout to input parameters
        self.main_layout = QGridLayout()
        self.setLayout(self.main_layout)

        # Create QComboBox for project
        row = 0
        self.main_layout.addWidget(QLabel("Project"), row, 0)
        self.project_combo_box = DynamicComboBox("configs/sys_projects.csv", title="Add/Delete Project")
        self.project_combo_box.setStatusTip("Select the project")
        self.main_layout.addWidget(self.project_combo_box, row, 1)

        # Create QComboBox for study case
        row += 1
        self.main_layout.addWidget(QLabel("Study"), row, 0)
        self.study_combo_box = DynamicComboBox("configs/sys_study.csv", title="Add/Delete Study")
        self.study_combo_box.setStatusTip("Select the study")
        self.main_layout.addWidget(self.study_combo_box, row, 1)

        # Create QComboBox for side
        row += 1
        self.main_layout.addWidget(QLabel("Side"), row, 0)
        self.side_combo_box = QComboBox()
        self.side_combo_box.addItems(sys_config.side)
        self.side_combo_box.setStatusTip("Select the subject side")
        self.main_layout.addWidget(self.side_combo_box, row, 1)

        # Create QComboBox for orientation
        row += 1
        self.main_layout.addWidget(QLabel("Orientation"), row, 0)
        self.orientation_combo_box = QComboBox()
        self.orientation_combo_box.addItems(sys_config.orientation)
        self.orientation_combo_box.setStatusTip("Select the subject orientation")
        self.main_layout.addWidget(self.orientation_combo_box, row, 1)

        # Create QLineEdit for user id
        row += 1
        self.main_layout.addWidget(QLabel("User"), row, 0)
        self.user_line_edit = QLineEdit()
        self.user_line_edit.setPlaceholderText("User")
        self.user_line_edit.setStatusTip("Write the user name")
        self.main_layout.addWidget(self.user_line_edit, row, 1)

        # Create QLineEdit for subject id
        row += 1
        date = datetime.now()
        date_string = date.strftime("%Y.%m.%d.%H.%M.%S")[:-3]
        self.main_layout.addWidget(QLabel("Subject ID"), row, 0)
        self.id_line_edit = QLineEdit()
        self.id_line_edit.setPlaceholderText(date_string)
        self.id_line_edit.setStatusTip("Write the subject id")
        self.main_layout.addWidget(self.id_line_edit, row, 1)

        # Create QLineEdit for study id
        row += 1
        date = datetime.now()
        self.main_layout.addWidget(QLabel("Study ID"), row, 0)
        self.idS_line_edit = QLineEdit()
        self.idS_line_edit.setPlaceholderText(date_string)
        self.idS_line_edit.setStatusTip("Write the study id")
        self.main_layout.addWidget(self.idS_line_edit, row, 1)

        # Create QLineEdit for subject name
        row += 1
        self.main_layout.addWidget(QLabel("Subject name"), row, 0)
        self.name_line_edit = QLineEdit()
        self.name_line_edit.setPlaceholderText("Name")
        self.name_line_edit.setStatusTip("Write the subject name")
        self.main_layout.addWidget(self.name_line_edit, row, 1)

        # Create QLineEdit for subject surname
        row += 1
        self.main_layout.addWidget(QLabel("Subject surname"), row, 0)
        self.surname_line_edit = QLineEdit()
        self.surname_line_edit.setPlaceholderText("Surname")
        self.surname_line_edit.setStatusTip("Write the subject surname")
        self.main_layout.addWidget(self.surname_line_edit, row, 1)

        # Create QLineEdit for subject birthday
        row += 1
        self.main_layout.addWidget(QLabel("Subject birthday"), row, 0)
        self.birthday_line_edit = QLineEdit()
        self.birthday_line_edit.setPlaceholderText("YY/MM/DD")
        self.birthday_line_edit.setStatusTip("Write the subject birthdate")
        self.main_layout.addWidget(self.birthday_line_edit, row, 1)

        # Create QLineEdit for subject wight
        row += 1
        self.main_layout.addWidget(QLabel("Subject weight"), row, 0)
        self.weight_line_edit = QLineEdit()
        self.weight_line_edit.setPlaceholderText("kg")
        self.weight_line_edit.setStatusTip("Write the subject weight")
        self.main_layout.addWidget(self.weight_line_edit, row, 1)

        # Create QLineEdit for subject height
        row += 1
        self.main_layout.addWidget(QLabel("Subject height"), row, 0)
        self.height_line_edit = QLineEdit()
        self.height_line_edit.setPlaceholderText("cm")
        self.height_line_edit.setStatusTip("Write the subject height")
        self.main_layout.addWidget(self.height_line_edit, row, 1)

        # Create QComboBox for RF coil
        row += 1
        self.main_layout.addWidget(QLabel("RF coil"), row, 0)
        self.rf_coil_combo_box = QComboBox()
        self.rf_coil_combo_box.setStatusTip("Select the rf coil")
        self.rf_coil_combo_box.addItems(hw_config.antenna_dict.keys())
        self.main_layout.addWidget(self.rf_coil_combo_box, row, 1)

        # Create QLineEdit for scanner
        row += 1
        self.main_layout.addWidget(QLabel("Scanner"), row, 0)
        self.scanner_line_edit = QLineEdit(hw_config.scanner_name)
        self.scanner_line_edit.setDisabled(True)
        self.scanner_line_edit.setStatusTip("Scanner version")
        self.main_layout.addWidget(self.scanner_line_edit, row, 1)

        # Create QLineEdit for software version
        # Get repo version
        try:
            tag = subprocess.check_output(['git', 'describe', '--tags'], stderr=subprocess.STDOUT).strip().decode(
                'utf-8')
        except subprocess.CalledProcessError as e:
            print(f"Error getting Git tag: {e.output.decode('utf-8')}")
            tag = ""
        row += 1
        self.main_layout.addWidget(QLabel("Software version"), row, 0)
        self.software_line_edit = QLineEdit(tag)
        self.software_line_edit.setDisabled(True)
        self.software_line_edit.setStatusTip("Software version")
        self.main_layout.addWidget(self.software_line_edit, row, 1)

        # Session dictionary
        self.session = {}

class DynamicComboBox(QComboBox):
    def __init__(self, file_name, title="Add/Delete Project"):
        super().__init__()
        self.file_name = file_name
        self.title = title

        # Load items from CSV
        self.load_items()

        # Add the special entry
        self.addItem("Add/Delete...")
        self.activated.connect(self.check_add_new)

    def check_add_new(self, index):
        if self.itemText(index) == "Add/Delete...":
            new_item, ok = QInputDialog.getText(self, self.title, "Enter new item:")

            if ok and new_item:
                existing_index = self.findText(new_item)

                if existing_index != -1:  # Item already exists → Remove it
                    self.removeItem(existing_index)
                    QMessageBox.information(self, "Item Removed", f"'{new_item}' has been removed.")
                else:  # Item does not exist → Add it before "Add/Delete..."
                    self.insertItem(self.count() - 1, new_item)
                    self.setCurrentIndex(self.count() - 2)  # Select the new item

                self.save_items()  # Update CSV file

    def load_items(self):
        """Load saved items from CSV file."""
        if os.path.exists(self.file_name):
            with open(self.file_name, newline='', encoding='utf-8') as file:
                reader = csv.reader(file)
                for row in reader:
                    if row:  # Avoid empty lines
                        self.addItem(row[0])

    def save_items(self):
        """Save current items to CSV file."""


        # Ensure the configs directory exists before writing
        os.makedirs(os.path.dirname(self.file_name), exist_ok=True)

        with open(self.file_name, "w", newline='', encoding='utf-8') as file:
            writer = csv.writer(file)
            for i in range(self.count()):
                item_text = self.itemText(i)
                if item_text != "Add/Delete...":  # Don't save the special entry
                    writer.writerow([item_text])


if __name__ == '__main__':
    # Only one QApplication for event loop
    app = QApplication(sys.argv)

    # Instantiate the window
    window = SessionWidget()
    window.show()

    # Start the event loop.
    app.exec()


from PyQt5.QtWidgets import QPushButton, QVBoxLayout, QLabel, QCheckBox, QLineEdit, QHBoxLayout, QGroupBox, QWidget


class PreProcessingTabWidget(QWidget):
    """
    PreProcessingTabWidget class for displaying a tab widget for image pre-processing options.

    Inherits from QTabWidget provided by PyQt5 to display a tab widget for image pre-processing options.

    Attributes:
        main: The parent widget.
    """

    def __init__(self, parent, *args, **kwargs):
        """
        Initialize the PreProcessingTabWidget.

        Args:
            parent: The parent widget.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(PreProcessingTabWidget, self).__init__(*args, **kwargs)

        # The 'main' attribute represents the parent widget, which is used to access the main window or controller.
        self.main = parent

        #***********************
        # Get single image
        #***********************
        self.scan_field = QLineEdit()
        self.scan_field.setText("All")
        self.scan_field.setStatusTip("'All' or integer number")
        self.scan_label = QLabel("scans")
        self.scan_button = QPushButton("Select")

        scan_layout_1 = QHBoxLayout()
        scan_layout_1.addWidget(self.scan_label)
        scan_layout_1.addWidget(self.scan_field)

        scan_layout = QVBoxLayout()
        scan_layout.addLayout(scan_layout_1)
        scan_layout.addWidget(self.scan_button)

        self.scan_group = QGroupBox("scan selection")
        self.scan_group.setLayout(scan_layout)

        #***********************
        # Cosbell
        #***********************
        #Checkboxes
        self.readout_checkbox = QCheckBox('Readout')
        self.phase_checkbox = QCheckBox('Phase')
        self.slice_checkbox = QCheckBox('Slice')
        self.cosbell_order_label = QLabel('Order')
        self.cosbell_order_field = QLineEdit()
        self.cosbell_order_field.setText('1')
        self.image_cosbell_button = QPushButton('Cosbell filter')

        # Cosbell layout
        self.checkbox_layout = QHBoxLayout()
        self.checkbox_layout.addWidget(self.readout_checkbox)
        self.checkbox_layout.addWidget(self.phase_checkbox)
        self.checkbox_layout.addWidget(self.slice_checkbox)

        # Order layout
        self.cosbell_order_layout = QHBoxLayout()
        self.cosbell_order_layout.addWidget(self.cosbell_order_label)
        self.cosbell_order_layout.addWidget(self.cosbell_order_field)

        self.cosbell_layout = QVBoxLayout()
        self.cosbell_layout.addLayout(self.checkbox_layout)
        self.cosbell_layout.addLayout(self.cosbell_order_layout)
        self.cosbell_layout.addWidget(self.image_cosbell_button)

        self.cosbell_group = QGroupBox("Cosbell")
        self.cosbell_group.setLayout(self.cosbell_layout)

        # ***********************
        # Zero Padding
        # ***********************
        self.zero_padding_order_label = QLabel('Matrix Size')
        self.zero_padding_order_field = QLineEdit()
        self.zero_padding_order_field.setPlaceholderText("Readout, Phase, Slice")
        self.zero_padding_order_field.setStatusTip('Must be integer numbers')

        self.zero_padding_order_layout = QHBoxLayout()
        self.zero_padding_order_layout.addWidget(self.zero_padding_order_label)
        self.zero_padding_order_layout.addWidget(self.zero_padding_order_field)

        self.image_padding_button = QPushButton('Zero padding')

        self.zero_padding_layout = QVBoxLayout()
        self.zero_padding_layout.addLayout(self.zero_padding_order_layout)
        self.zero_padding_layout.addWidget(self.image_padding_button)

        self.zero_padding_group = QGroupBox("Zero Padding")
        self.zero_padding_group.setLayout(self.zero_padding_layout)

        # ***********************
        # FOV change
        # ***********************
        self.change_fov_label = QLabel('Shift (mm)')
        self.change_fov_field = QLineEdit()
        self.change_fov_field.setPlaceholderText("Readout, Phase, Slice")

        self.change_fov_layout = QHBoxLayout()
        self.change_fov_layout.addWidget(self.change_fov_label)
        self.change_fov_layout.addWidget(self.change_fov_field)

        self.new_fov_button = QPushButton('Shift FOV')

        self.new_fov_layout = QVBoxLayout()
        self.new_fov_layout.addLayout(self.change_fov_layout)
        self.new_fov_layout.addWidget(self.new_fov_button)

        self.new_fov_group = QGroupBox("FOV Shift")
        self.new_fov_group.setLayout(self.new_fov_layout)

        # Main layout
        self.preprocessing_layout = QVBoxLayout()
        self.preprocessing_layout.addWidget(self.scan_group)
        self.preprocessing_layout.addWidget(self.zero_padding_group)
        self.preprocessing_layout.addWidget(self.cosbell_group)
        self.preprocessing_layout.addWidget(self.new_fov_group)
        self.preprocessing_layout.addStretch()
        self.setLayout(self.preprocessing_layout)



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import sys
import threading

from PyQt5.QtCore import QEvent

from marge.seq.sequences import defaultsequences
from marge.ui.window_main import MainWindow


class MainController(MainWindow):
    def __init__(self, *args, **kwargs):
        super(MainController, self).__init__(*args, **kwargs)

        self.set_session(self.session)

        self.initializeThread()

    def set_demo(self, demo):
        self.demo = demo

    def set_session(self, session):
        # Set window title
        self.session = session
        self.setWindowTitle("MaRGE " + session["software_version"] + ": " + session['directory'])
        # Add the session to all sequences
        for sequence in defaultsequences.values():
            sequence.session = session

    def initializeThread(self):
        # Start the sniffer
        thread = threading.Thread(target=self.history_list.waitingForRun)
        thread.start()
        print("Sniffer initialized.\n")

    def set_console(self):
        self.layout_left.addWidget(self.console)

    def changeEvent(self, event):
        if event.type() == QEvent.ActivationChange:  # Event type 99
            if self.isActiveWindow():
                self.set_console()
        super().changeEvent(event)

    def closeEvent(self, event):
        """
        Shuts down the application on close.

        This method is called when the application is being closed. It sets the `app_open` flag to False, restores
        `sys.stdout` to its default value, and performs additional cleanup tasks if the `demo` flag is not set.
        It also prints a closing message to the console.

        Args:
            event (QCloseEvent): The close event triggered by the user.

        Returns:
            None
        """
        # Return stdout to defaults.
        sys.stdout = sys.__stdout__
            
        print('\nMain GUI closed successfully!')

        self.parent.show()

        super().closeEvent(event)



"""
:author:    José Miguel Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import imageio.v2 as imageio

from marge.controller.controller_plot3d import Plot3DController as Spectrum3DPlot
from marge.widgets.widget_figures import FiguresLayoutWidget


class FiguresLayoutController(FiguresLayoutWidget):
    """
    Controller for figures layout
    """
    def __init__(self, main):
        super().__init__()
        self.main = main

        # Show the initial plot
        self.firstPlot()

        # Show the wellcome message
        self.wellcomeMessage()

    @staticmethod
    def wellcomeMessage():
        """
        Display the welcome message.

        This static method displays the welcome message for the Graphical User Interface for MaRCoS. It prints information about the developer, contact details, and the GitHub repository URL.

        Note:
            The method does not return any value.
        """
        print("Graphical User Interface for MaRCoS")
        print("J.M. Algarín, PhD")
        print("josalggui@i3m.upv.es")
        print("mriLab @ i3M, CSIC, Valencia, Spain")
        print("https://github.com/mriLab-i3M/MaRGE\n")

    def firstPlot(self):
        """
        Display the first plot.

        This method displays the first plot by loading an image, creating a Spectrum3DPlot object, and customizing its appearance. The plot is added to the figures layout.

        Note:
            The method does not return any value.
        """
        try:
            logo = imageio.imread("resources/images/logo.png")
        except:
            logo = imageio.imread("../../resources/images/logo.png")
        self.clearFiguresLayout()
        welcome = Spectrum3DPlot(main=self, data=logo.transpose([1, 0, 2]),
                                 title='Institute for Instrumentation in Molecular Imaging (i3M)')
        welcome.hideAxis('bottom')
        welcome.hideAxis('left')
        welcome.showHistogram(False)
        welcome.ui.menuBtn.hide()
        welcome.ui.roiBtn.hide()
        self.addWidget(welcome)

    def clearFiguresLayout(self) -> None:
        """
        Clear the figures layout.

        This method removes all widgets from the figures layout.

        Returns:
            None
        """
        for ii in range(self.layout.count()):
            item = self.layout.takeAt(0)
            item.widget().deleteLater()



#  NanoVNASaver
#
#  A python program to view and export Touchstone data from a NanoVNA
#  Copyright (C) 2019, 2020  Rune B. Broberg
#  Copyright (C) 2020,2021 NanoVNA-Saver Authors
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
import platform
from collections import namedtuple
from time import sleep
from typing import List

import serial
from serial.tools import list_ports
from serial.tools.list_ports_common import ListPortInfo

from marge.vna.VNA import VNA
from marge.vna.AVNA import AVNA
from marge.vna.NanoVNA import NanoVNA
from marge.vna.NanoVNA_F import NanoVNA_F
from marge.vna.NanoVNA_F_V2 import NanoVNA_F_V2
from marge.vna.NanoVNA_H import NanoVNA_H
from marge.vna.NanoVNA_H4 import NanoVNA_H4
from marge.vna.NanoVNA_V2 import NanoVNA_V2
from marge.vna.TinySA import TinySA
from marge.vna.Serial import drain_serial, Interface

logger = logging.getLogger(__name__)

USBDevice = namedtuple("Device", "vid pid name")

USBDEVICETYPES = (
    USBDevice(0x0483, 0x5740, "NanoVNA"),
    USBDevice(0x16c0, 0x0483, "AVNA"),
    USBDevice(0x04b4, 0x0008, "S-A-A-2"),
)
RETRIES = 3
TIMEOUT = 0.2
WAIT = 0.05

NAME2DEVICE = {
    "S-A-A-2": NanoVNA_V2,
    "AVNA": AVNA,
    "H4": NanoVNA_H4,
    "H": NanoVNA_H,
    "F_V2": NanoVNA_F_V2,
    "F": NanoVNA_F,
    "NanoVNA": NanoVNA,
    "tinySA": TinySA,
    "Unknown": NanoVNA,
}


# The USB Driver for NanoVNA V2 seems to deliver an
# incompatible hardware info like:
# 'PORTS\\VID_04B4&PID_0008\\DEMO'
# This function will fix it.


def _fix_v2_hwinfo(dev):
    # if dev.hwid == r'PORTS\VID_04B4&PID_0008\DEMO':
    if r'PORTS\VID_04B4&PID_0008' in dev.hwid:
        dev.vid, dev.pid = 0x04b4, 0x0008
    return dev


def usb_typename(device: ListPortInfo) -> str:
    return next((t.name for t in USBDEVICETYPES if
                 device.vid == t.vid and device.pid == t.pid),
                "")

# Get list of interfaces with VNAs connected


def get_interfaces() -> List[Interface]:
    interfaces = []
    # serial like usb interfaces
    for d in list_ports.comports():
        if platform.system() == 'Windows' and d.vid is None:
            d = _fix_v2_hwinfo(d)
        if not (typename := usb_typename(d)):
            continue
        logger.debug("Found %s USB:(%04x:%04x) on port %s",
                     typename, d.vid, d.pid, d.device)
        iface = Interface('serial', typename)
        iface.port = d.device
        iface.open()
        iface.comment = get_comment(iface)
        iface.close()
        interfaces.append(iface)

    logger.debug("Interfaces: %s", interfaces)
    return interfaces


def get_portinfos() -> List[str]:
    portinfos = []
    # serial like usb interfaces
    for d in list_ports.comports():
        logger.debug("Found USB:(%04x:%04x) on port %s",
                     d.vid, d.pid, d.device)
        iface = Interface('serial', "DEBUG")
        iface.port = d.device
        iface.open()
        version = detect_version(iface)
        iface.close()
        portinfos.append(version)
    return portinfos


def get_VNA(iface: Interface) -> VNA:
    # serial_port.timeout = TIMEOUT
    return NAME2DEVICE[iface.comment](iface)


def get_comment(iface: Interface) -> str:
    logger.info("Finding correct VNA type...")
    with iface.lock:
        vna_version = detect_version(iface)

    if vna_version == 'v2':
        return "S-A-A-2"

    logger.info("Finding firmware variant...")
    info = get_info(iface)
    for search, name in (
            ("AVNA + Teensy", "AVNA"),
            ("NanoVNA-H 4", "H4"),
            ("NanoVNA-H", "H"),
            ("NanoVNA-F_V2", "F_V2"),
            ("NanoVNA-F", "F"),
            ("NanoVNA", "NanoVNA"),
            ("tinySA", "tinySA"),
    ):
        if info.find(search) >= 0:
            return name
    logger.warning("Did not recognize NanoVNA type from firmware.")
    return "Unknown"


def detect_version(serial_port: serial.Serial) -> str:
    data = ""
    for i in range(RETRIES):
        drain_serial(serial_port)
        serial_port.write("\r".encode("ascii"))
        # workaround for some UnicodeDecodeError ... repeat ;-)
        drain_serial(serial_port)
        serial_port.write("\r".encode("ascii"))
        sleep(0.05)

        data = serial_port.read(128).decode("ascii")
        if data.startswith("ch> "):
            return "v1"
        # -H versions
        if data.startswith("\r\nch> "):
            return "vh"
        if data.startswith("\r\n?\r\nch> "):
            return "vh"
        if data.startswith("2"):
            return "v2"
        logger.debug("Retry detection: %s", i + 1)
    logger.error('No VNA detected. Hardware responded to CR with: %s', data)
    return ""


def get_info(serial_port: serial.Serial) -> str:
    for _ in range(RETRIES):
        drain_serial(serial_port)
        serial_port.write("info\r".encode("ascii"))
        lines = []
        retries = 0
        while True:
            line = serial_port.readline()
            line = line.decode("ascii").strip()
            if not line:
                retries += 1
                if retries > RETRIES:
                    return ""
                sleep(WAIT)
                continue
            if line == "info":  # suppress echo
                continue
            if line.startswith("ch>"):
                logger.debug("Needed retries: %s", retries)
                break
            lines.append(line)
        logger.debug("Info output: %s", lines)
        return "".join(lines)



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
from PyQt5.QtWidgets import QSizePolicy
from pyqtgraph import LayoutWidget


class FiguresLayoutWidget(LayoutWidget):
    def __init__(self):
        super().__init__()



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
import numpy as np
import pyqtgraph as pg


class Plot3DWidget(pg.ImageView):
    def __init__(self, main):

        # Save inputs into the self
        self.main = main

        # Define the PlotItem to display the image
        self.plot_item = pg.PlotItem()

        # Create text item
        self.text_item = pg.TextItem()
        self.text_item.setText('', color='red')

        # Execute the parent init
        super(Plot3DWidget, self).__init__(view=self.plot_item)

        # Create ROI to get FOV
        self.roiFOV = pg.ROI([0, 0], [1, 1])
        self.view.addItem(self.roiFOV)

        # Insert text_item into the ImageViewer widget
        self.vbox = self.getView()
        self.vbox.addItem(self.text_item)



#!/usr/bin/env python3
#
# Integrated test of the MaRCoS server, HDL and compiler.
#
# Script compiles and runs the MaRCoS Verilator simulation (server +
# HDL model) and sends it various binaries generated using the MaRCoS
# client compiler, then compares the simulated hardware output against
# the expected output.
#
# To run a single test, use e.g.:
# python -m unittest test_marga_model.Modeltest.test_many_quick

from test_base import *

class ModelTest(unittest.TestCase):
    """Main test class for general HDL and compiler development/debugging;
    inputs to the API and UUT are either a CSV file or a dictionary,
    output is another CSV file, and comparison is either between the
    input and output CSV files (with allowance made for memory/startup
    delays) or between the output CSV and a reference CSV file. If a
    dictionary is used as input, a reference CSV file must be created.
    """

    @classmethod
    def setUpClass(cls):
        # TODO make this check for a file first
        subprocess.call(["make", "-j4", "-s", "-C", os.path.join(marga_sim_path, "build")])
        subprocess.call(["fallocate", "-l", "516KiB", "/tmp/marcos_server_mem"])
        subprocess.call(["killall", "marga_sim"], stderr=subprocess.DEVNULL) # in case other instances were started earlier

        warnings.simplefilter("ignore", mc.MarServerWarning)

    def setUp(self):
        # start simulation
        if marga_sim_fst_dump:
            self.p = subprocess.Popen([os.path.join(marga_sim_path, "build", "marga_sim"), "both", marga_sim_csv, marga_sim_fst],
                                      stdout=subprocess.DEVNULL,
                                      stderr=subprocess.STDOUT)
        else:
            self.p = subprocess.Popen([os.path.join(marga_sim_path, "build", "marga_sim"), "csv", marga_sim_csv],
                                      stdout=subprocess.DEVNULL,
                                      stderr=subprocess.STDOUT)


        # open socket
        time.sleep(0.05) # give marga_sim time to start up

        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.connect((ip_address, port)) # only connect to local simulator
        self.packet_idx = 0

    def tearDown(self):
        # self.p.terminate() # if not already terminated
        # self.p.kill() # if not already terminated
        self.s.close()

        if marga_sim_fst_dump:
            # open GTKWave
            os.system("gtkwave " + marga_sim_fst + " " + os.path.join(marga_sim_path, "src", "marga_sim.sav"))

    ## Tests are approximately in order of complexity

    def test_single(self):
        """ Basic state change on a single buffer """
        refl, siml = compare_csv("test_single", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_four_par(self):
        """ State change on four buffers in parallel """
        refl, siml = compare_csv("test_four_par", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_long_time(self):
        """ State change on four buffers in parallel """
        max_orig = mc.COUNTER_MAX
        mc.COUNTER_MAX = 0xfff # temporarily reduce max time used by compiler
        refl, siml = compare_csv("test_long_time", self.s, self.p)
        mc.COUNTER_MAX = max_orig
        self.assertEqual(refl, siml)

    def test_single_quick(self):
        """ Quick successive state changes on a single buffer 1 cycle apart """
        refl, siml = compare_csv("test_single_quick", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_single_delays(self):
        """ State changes on a single buffer with various delays in between"""
        refl, siml = compare_csv("test_single_delays", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_two_quick(self):
        """ Quick successive state changes on two buffers, one cycle apart """
        refl, siml = compare_csv("test_two_quick", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_two_delays(self):
        """ State changes on two buffers, various delays in between """
        refl, siml = compare_csv("test_two_delays", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_three_quick(self):
        """ Quick successive state changes on two buffers, one cycle apart """
        refl, siml = compare_csv("test_three_quick", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_three_delays(self):
        """ Successive state changes on three buffers, two cycles apart """
        refl, siml = compare_csv("test_three_delays", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_mult_quick(self):
        """ Quick successive state changes on multiple buffers, 1 cycle apart """
        refl, siml = compare_csv("test_mult_quick", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_many_quick(self):
        """ Many quick successive state changes on multiple buffers, all 1 cycle apart """
        refl, siml = compare_csv("test_many_quick", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_stream_quick(self):
        """ Bursts of state changes on multiple buffers with uneven gaps for each individual buffer, each state change 1 cycle apart """
        refl, siml = compare_csv("test_stream_quick", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_uneven_times(self):
        """ Bursts of state changes on multiple buffers with uneven gaps, each state change uneven numbers of cycles apart """
        refl, siml = compare_csv("test_uneven_times", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_uneven_sparse(self):
        """ Bursts of state changes on multiple buffers with uneven gaps, each state change uneven numbers of cycles apart """
        refl, siml = compare_csv("test_uneven_sparse", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_cfg(self):
        """ Configuration and LED bits/words """
        refl, siml = compare_csv("test_cfg", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_rx_simple(self):
        """ RX window with realistic RX rate configuration, resetting and gating """
        refl, siml = compare_csv("test_rx_simple", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_two_uneven_latencies(self):
        """Simultaneous state changes on two buffers, however 2nd buffer is
        specified to have 1 cycle of latency more than 1st - so its changes
        occur earlier to compensate"""
        refl, siml = compare_csv("test_two_uneven_latencies", self.s, self.p,
                                  latencies=np.array([
                                      0,0,0,0,
                                      0,0,1,0,
                                      0,0,0,0,
                                      0,0,0,0,
                                      0], dtype=np.uint16),
                                  self_ref=False)
        self.assertEqual(refl, siml)

    def test_many_uneven_latencies(self):
        """Simultaneous state changes on four buffers, however they are
        assumed to all have different latencies relative to each other - thus
        out-of-sync change requests turn out in sync"""
        refl, siml = compare_csv("test_many_uneven_latencies", self.s, self.p,
                                  latencies=np.array([
                                      0, 0, 0, # grad
                                      0, 0, # rx
                                      2, 4, 6, 8, # tx
                                      0, 0, 0, 0, 0, 0, # lo phase
                                      0, 0 # gates and LEDs, RX config
                                  ], dtype=np.uint16),
                                  self_ref=False)
        self.assertEqual(refl, siml)

    def test_fhd_single(self):
        """Single state change on GPA-FHDO x gradient output, default SPI
        clock divisor; simultaneous change on TX0i"""
        set_grad_board("gpa-fhdo")
        refl, siml = compare_csv("test_fhd_single", self.s, self.p, **fhd_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_fhd_series(self):
        """Series of state changes on GPA-FHDO x gradient output, default SPI
        clock divisor """
        set_grad_board("gpa-fhdo")
        refl, siml = compare_csv("test_fhd_series", self.s, self.p, **fhd_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_fhd_multiple(self):
        """A few state changes on GPA-FHDO gradient outputs, default SPI clock divisor"""
        set_grad_board("gpa-fhdo")
        refl, siml = compare_csv("test_fhd_multiple", self.s, self.p, **fhd_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_fhd_many(self):
        """Many state changes on GPA-FHDO gradient outputs, default SPI clock divisor - simultaneous with similar TX changes"""
        set_grad_board("gpa-fhdo")
        refl, siml = compare_csv("test_fhd_many", self.s, self.p, **fhd_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_fhd_too_fast(self):
        """Two state changes on GPA-FHDO gradient outputs, default SPI clock
        divisor - too fast for the divider, second state change won't
        be applied and server should notice the error
        """
        set_grad_board("gpa-fhdo")

        # Run twice, to catch two different warnings (I couldn't find a more straightforward way to do this)
        with self.assertWarns( RuntimeWarning, msg="expected gpa-fhdo error not observed") as cmr:
            refl, siml = compare_csv("test_fhd_too_fast", self.s, self.p, self_ref=False, **fhd_config)
        with self.assertWarns( UserWarning, msg="expected marcompile warning not observed") as cmu:
            self.tearDown()
            self.setUp()
            refl, siml = compare_csv("test_fhd_too_fast", self.s, self.p, self_ref=False, **fhd_config)

        restore_grad_board()
        # self.assertEqual( str(cm.exception) , "gpa-fhdo gradient error; possibly missing samples")
        self.assertEqual( str(cmu.warning), "Gradient updates are too frequent for selected SPI divider. Missed samples are likely!")
        self.assertEqual( str(cmr.warning) , "SERVER ERROR: gpa-fhdo gradient error; possibly missing samples")
        self.assertEqual(refl, siml)

    def test_oc1_single(self):
        """Single state change on ocra1 x gradient output, default SPI clock
        divisor; simultaneous change on TX0i
        """
        set_grad_board("ocra1")
        refl, siml = compare_csv("test_oc1_single", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_oc1_series(self):
        """Series of state changes on ocra1 x gradient output, default SPI
        clock divisor
        """
        set_grad_board("ocra1")
        refl, siml = compare_csv("test_oc1_series", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_oc1_two(self):
        """Two sets of simultaneous state changes on ocra1 gradient outputs,
        default SPI clock divisor
        """
        set_grad_board("ocra1")
        refl, siml = compare_csv("test_oc1_two", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_oc1_two_same(self):
        """One set of simultaneous identical state changes on ocra1 gradient
        outputs, then a single change later; default SPI clock divisor
        """
        set_grad_board("ocra1")
        refl, siml = compare_csv("test_oc1_two_same", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_oc1_four(self):
        """Four simultaneous state changes on ocra1 gradient outputs, default
        SPI clock divisor
        """
        set_grad_board("ocra1")
        refl, siml = compare_csv("test_oc1_four", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_oc1_series_same(self):
        """Several sets of simultaneous state changes on ocra1 gradient
        outputs, with MSBs and LSBs kept similar in a pattern to test
        a bug; default SPI clock divisor
        """
        set_grad_board("ocra1")
        refl, siml = compare_csv("test_oc1_series_same", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)
        # self.assertEqual(1, 0)

    def test_oc1_many(self):
        """Multiple simultaneous state changes on ocra1 gradient outputs, default
        SPI clock divisor
        """
        set_grad_board("ocra1")
        refl, siml = compare_csv("test_oc1_many", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_oc1_too_fast(self):
        """Two state changes on ocra1 gradient outputs, default SPI clock
        divisor - too fast for the divider, second state change won't
        be applied and server should notice the error
        """
        set_grad_board("ocra1")

        # Run twice, to catch two different warnings (I couldn't find a more straightforward way to do this)
        with self.assertWarns( RuntimeWarning, msg="expected ocra1 error not observed") as cmr:
            refl, siml = compare_csv("test_oc1_too_fast", self.s, self.p, self_ref=False, **oc1_config)
        with self.assertWarns( UserWarning, msg="expected marcompile warning not observed") as cmu:
            self.tearDown()
            self.setUp()
            refl, siml = compare_csv("test_oc1_too_fast", self.s, self.p, self_ref=False, **oc1_config)
            restore_grad_board()
        # self.assertEqual( str(cm.exception) , "gpa-fhdo gradient error; possibly missing samples")
        self.assertEqual( str(cmu.warning), "Gradient updates are too frequent for selected SPI divider. Missed samples are likely!")
        self.assertEqual( str(cmr.warning) , "SERVER ERROR: ocra1 gradient error; possibly missing samples")
        self.assertEqual(refl, siml)

    def test_single_dict(self):
        """ Basic state change on a single buffer. Dict version"""
        d = {'tx0_i': (np.array([100]), np.array([10000]))}
        refl, siml = compare_dict(d, "test_single", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_removed_instructions_dict(self):
        """Basic state change on a single buffer, with many more specified
        events than actual changes, leading to removed instructions and a
        series of warnings. Dict version"""
        reps = 2000
        d = {'tx0_i': ( np.arange(100, 100 + reps) , np.array([10000]*reps) )}
        with self.assertWarns( mc.MarRemovedInstructionWarning,
                               msg="expected marcompile warning not observed") as cmu:
            refl, siml = compare_dict(d, "test_single", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_four_par_dict(self):
        """ State change on four buffers in parallel. Dict version"""
        d = {'tx0_i': (np.array([100]), np.array([5000])),
             'tx0_q': (np.array([100]), np.array([15000])),
             'tx1_i': (np.array([100]), np.array([25000])),
             'tx1_q': (np.array([100]), np.array([35000]))}
        refl, siml = compare_dict(d, "test_four_par", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_uneven_sparse_dict(self):
        """ Bursts of state changes on multiple buffers with uneven gaps, each state change uneven numbers of cycles apart. Dict version"""
        d = {'tx0_i': (np.array([100, 110, 113, 114, 115, 130, 152, 153, 156, 170, 174, 300, 10000]),
                       np.array([1, 5, 9, 13, 17, 25, 33, 37, 41, 45, 49, 53, 57])),
             'tx0_q': (np.array([110, 113, 114, 115, 116, 152, 156, 170, 174, 300, 10001]),
                       np.array([6, 10, 14, 18, 22, 34, 42, 46, 50, 54, 62])),
             'tx1_i': (np.array([113, 114, 115, 116, 130, 152, 153, 170, 177, 300, 10001]),
                       np.array([11, 15, 19, 23, 27, 35, 39, 47, 51, 55, 63])),
             'tx1_q': (np.array([114, 115, 116, 130, 150, 152, 156, 170, 177, 300, 10000]),
                       np.array([16, 20, 24, 28, 32, 36, 44, 48, 52, 56, 60])) }
        refl, siml = compare_dict(d, "test_uneven_sparse", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_cfg_dict(self):
        """ Configuration and LED bits/words. Dict version"""
        d = {'rx0_rate': ( np.array([100, 110, 228, 230]), np.array([1, 0, 1, 0]) ),
             'rx1_rate': ( np.array([110, 120, 225, 228]), np.array([1, 0, 1, 0]) ),
             'rx0_rate_valid': ( np.array([120, 130, 222, 225]), np.array([1, 0, 1, 0]) ),
             'rx1_rate_valid': ( np.array([130, 140, 220, 222]), np.array([1, 0, 1, 0]) ),
             'rx0_rst_n': ( np.array([140, 150, 218, 220]), np.array([1, 0, 1, 0]) ),
             'rx1_rst_n': ( np.array([150, 160, 216, 218]), np.array([1, 0, 1, 0]) ),
             'rx0_en': ( np.array([160, 170, 215, 216]), np.array([1, 0, 1, 0]) ),
             'rx1_en': ( np.array([170, 180, 214, 215]), np.array([1, 0, 1, 0]) ),
             'tx_gate': ( np.array([180, 190, 213, 214]), np.array([1, 0, 1, 0]) ),
             'rx_gate': ( np.array([190, 200, 212, 213]), np.array([1, 0, 1, 0]) ),
             'trig_out': ( np.array([200, 210, 211, 212]), np.array([1, 0, 1, 0]) ),
             'leds': ( np.array([100, 110, 120, 130, 140, 150, 160, 170, 180, 190, 200, 210,
                                 211, 212, 213, 214, 215, 216, 218, 220, 222, 225, 228, 230]),
                       np.array([  1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11, 12,
                                   20,  30,  40,  50,  60,  70,  80,  90, 100, 110, 120, 255]) )
             }
        refl, siml = compare_dict(d, "test_cfg", self.s, self.p)
        self.assertEqual(refl, siml)

    def test_fhd_many_dict(self):
        """Many state changes on GPA-FHDO gradient outputs, default SPI clock divisor - simultaneous with similar TX changes. Dict version"""
        set_grad_board("gpa-fhdo")
        d = {'tx0_i': (np.array([600, 1800]), np.array([1, 65532])),
             'tx0_q': (np.array([900, 2100]), np.array([2, 65533])),
             'tx1_i': (np.array([1200, 2400]), np.array([3, 65534])),
             'tx1_q': (np.array([1500, 2700]), np.array([4, 65535])),
             'fhdo_vx': (np.array([600, 1800]), np.array([1, 65532])),
             'fhdo_vy': (np.array([900, 2100]), np.array([2, 65533])),
             'fhdo_vz': (np.array([1200, 2400]), np.array([3, 65534])),
             'fhdo_vz2': (np.array([1500, 2700]), np.array([4, 65535]))}
        refl, siml = compare_dict(d, "test_fhd_many", self.s, self.p, **fhd_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_oc1_many_dict(self):
        """Many state changes on OCRA1 gradient outputs, default SPI clock divisor - simultaneous with similar TX changes. Dict version"""
        set_grad_board("ocra1")
        d = {'tx0_i': (np.array([600, 900, 1200, 1500, 1800, 2100, 2400, 2700]),
                          np.array([1, 5, 9, 13, 65535, 65534, 10000, 20000])),
             'tx0_q': (np.array([600, 1200, 1500, 1800, 2100, 2400, 2700]),
                          np.array([2, 10, 14, 2, 65533, 12000, 25000])),
             'tx1_i': (np.array([600, 900, 1500, 1800, 2100, 2400, 2700]),
                          np.array([3, 7, 15, 3, 65532, 14000, 30000])),
             'tx1_q': (np.array([600, 900, 1200, 1500, 1800, 2100, 2400, 2700]),
                          np.array([4, 8, 12, 16, 4, 65531, 16000, 35000])),

             'ocra1_vx': (np.array([600, 900, 1200, 1500, 1800, 2100, 2400, 2700]),
                          np.array([1, 5, 9, 13, 262143, 262142, 100000, 20000])),
             'ocra1_vy': (np.array([600, 1200, 1500, 1800, 2100, 2400, 2700]),
                          np.array([2, 10, 14, 2, 262141, 120000, 25000])),
             'ocra1_vz': (np.array([600, 900, 1500, 1800, 2100, 2400, 2700]),
                          np.array([3, 7, 15, 3, 262140, 140000, 30000])),
             'ocra1_vz2': (np.array([600, 900, 1200, 1500, 1800, 2100, 2400, 2700]),
                          np.array([4, 8, 12, 16, 4, 262139, 160000, 35000])),
             }
        refl, siml = compare_dict(d, "test_oc1_many", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_oc1_two_dict(self):
        """Many state changes on OCRA1 gradient outputs, default SPI clock divisor - attempt to probe dict sorting bug."""
        set_grad_board("ocra1")
        d = {'ocra1_vx': (np.array([600, 900]), np.array([0x10001, 0x20002])),
             'ocra1_vy': (np.array([600]), np.array([0x10001]))}
        refl, siml = compare_dict(d, "test_oc1_two_dict", self.s, self.p, **oc1_config)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_single_expt(self):
        """ Basic state change on a single buffer. Experiment version"""
        set_grad_board("gpa-fhdo")
        d = {'tx0_i': (np.array([1]), np.array([0.5]))}
        expt_args = {'rx_t': 2, 'auto_leds': False}
        refl, siml = compare_expt_dict(d, "test_single_expt", self.s, self.p, **expt_args)
        self.assertEqual(refl, siml)

        # test for the other grad board
        self.tearDown(); self.setUp()
        set_grad_board("ocra1")
        refl, siml = compare_expt_dict(d, "test_single_expt", self.s, self.p, **expt_args)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_four_par_expt_iq(self):
        """ State change on four buffers in parallel. Experiment version using complex inputs"""
        d = {'tx0': (np.array([1]), np.array([0.5+0.2j])), 'tx1': (np.array([1]), np.array([-0.3+1j]))}
        expt_args = {'rx_t': 2, 'auto_leds': False}
        set_grad_board("gpa-fhdo")
        refl, siml = compare_expt_dict(d, "test_four_par_expt_iq", self.s, self.p, **expt_args)
        self.assertEqual(refl, siml)

        self.tearDown(); self.setUp()
        set_grad_board("ocra1")
        refl, siml = compare_expt_dict(d, "test_four_par_expt_iq", self.s, self.p, **expt_args)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_uneven_sparse_expt_fhd(self):
        """ Miscellaneous pulses on TX and gradients, with various acquisition windows """
        d = {'tx0': (np.array([10,15, 30,35, 100,105]), np.array([1,0, 0.8j,0, 0.7+0.2j,0])),
             'tx1': (np.array([5,20,  50,70,  110,125]), np.array([-1j,0,  -0.5j,0,  0.5+0.3j,0])),

             # gradient events may occur only on a single channel at a
             # time (i.e. no parallel updates) and must be spaced by
             # at least 4us - i.e. after any update occurs, the next
             # update (on any channel) may only occur 4us later
             'grad_vx': (np.array([ 10, 30, 54, 75, 100]), np.array([-1, 1, 0.5, -0.5, 0])),
             'grad_vy': (np.array([ 14, 26, 50, 71,  96]), np.array([-1, 1, 0.5, -0.5, 0])),
             'grad_vz': (np.array([ 6,  22,     79,  88]), np.array([-1, 1,       0.5, 0])),
             'grad_vz2': (np.array([18,     46, 67,  92]), np.array([-1,      1,  0.5, 0])),

             'rx0_en': (np.array([7,12,  30,40,  80,90]), np.array([1,0, 1,0, 1,0])),
             'rx1_en': (np.array([8,14,  33,45,  83,95]), np.array([1,0, 1,0, 1,0])),
             }

        set_grad_board("gpa-fhdo")
        expt_args = {'rx_t': 0.5, 'auto_leds': False}
        refl, siml = compare_expt_dict(d, "test_uneven_sparse_expt_fhd", self.s, self.p, **expt_args)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_uneven_sparse_expt_oc1(self):
        """ Miscellaneous pulses on TX and gradients, with various acquisition windows """
        d = {'tx0': (np.array([10,15, 30,35, 100,105]), np.array([1,0, 0.8j,0, 0.7+0.2j,0])),
             'tx1': (np.array([5,20,  50,70,  110,125]), np.array([-1j,0,  -0.5j,0,  0.5+0.3j,0])),

             # gradient events must be spaced by at least 4us -
             # i.e. after any update occurs, the next update (on any
             # channel) may only occur 4us later
             'grad_vx': (np.array([10, 30, 43, 50, 77]), np.array([-1, 1, 0.5, -0.5, 0])),
             'grad_vy': (np.array([15, 23, 43, 57, 84]), np.array([-1, 1, 0.5, -0.5, 0])),
             'grad_vz': (np.array([15, 23,     65, 77]), np.array([-1, 1, 0.5, -0.5, 0])),
             'grad_vz2': (np.array([15,     43, 65, 77]), np.array([-1, 1, 0.5, -0.5, 0])),

             'rx0_en': (np.array([7,12,  30,40,  80,90]), np.array([1,0, 1,0, 1,0])),
             'rx1_en': (np.array([8,14,  33,45,  83,95]), np.array([1,0, 1,0, 1,0])),
             }

        set_grad_board("ocra1")
        expt_args = {'rx_t': 0.5, 'auto_leds': False}
        refl, siml = compare_expt_dict(d, "test_uneven_sparse_expt_oc1", self.s, self.p, **expt_args)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_cic_shift_expt(self):
        """ Test the experiment code to program the open-source (non-Xilinx) variant of the CIC filter """
        # d = {'tx0': ( np.array([10, 15]), np.array([0.5, 0]) )}
        set_grad_board("ocra1")
        expt_args = {'rx_t': 0.5, 'set_cic_shift': True, 'auto_leds': False}
        refl, siml = compare_expt_dict({}, "test_cic_shift_expt", self.s, self.p, **expt_args)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_init_grad_expt_fhd(self):
        """ Test whether GPA-FHDO gradient events happening at time = 0 cause errors -- they should not if there's a sufficient initial wait"""
        set_grad_board("gpa-fhdo")
        expt_args = {'rx_t': 0.5, 'grad_max_update_rate': 0.1, 'auto_leds': False} # deliberately slow update rate
        d = {'grad_vx': (np.array([0, 5]), np.array([0.5, 0]))}
        refl, siml = compare_expt_dict(d, "test_init_grad_expt_fhd", self.s, self.p, **expt_args)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_init_grad_expt_oc1(self):
        """ Test whether OCRA1 gradient events happening at time = 0 cause errors -- they should not if there's a sufficient initial wait."""
        set_grad_board("ocra1")
        expt_args = {'rx_t': 0.5, 'grad_max_update_rate': 0.1, 'auto_leds': False} # deliberately slow update rate
        d = {'grad_vx': (np.array([0, 20]), np.array([0.5, 0])) }
        refl, siml = compare_expt_dict(d, "test_init_grad_expt_oc1", self.s, self.p, **expt_args)
        restore_grad_board()
        self.assertEqual(refl, siml)

    def test_auto_leds_expt(self):
        """ Test whether the auto-LED scan works correctly """
        expt_args = {'auto_leds': True}
        d = {'tx0_i': (np.array([0, 100]), np.array([1, 0])) }
        refl, siml = compare_expt_dict(d, "test_auto_leds_expt", self.s, self.p, **expt_args)
        self.assertEqual(refl, siml)

    def test_tx_complex_expt(self):
        """ Test whether the Experiment class removes duplicate i/q entries for complex TX data """
        max_rem_instr = mc.max_removed_instructions
        mc.max_removed_instructions = 1
        expt_args = {'auto_leds': False}
        d = {'tx0': (np.array([0, 10, 15, 20, 30, 50]), np.array([1-1j, 1+1j, 1j, -1+1j, -1, -1-1j]))}
        with warnings.catch_warnings():
            warnings.filterwarnings("error", category=mc.MarRemovedInstructionWarning)
            refl, siml = compare_expt_dict(d, "test_tx_complex_expt", self.s, self.p, **expt_args)

        # restore to default
        mc.max_removed_instructions = max_rem_instr
        self.assertEqual(refl, siml)

    def test_lo_change_expt(self):
        """ Test whether the Experiment class can handle changes in LO frequency, followed by being rerun """
        expt_args = {'lo_freq': 1, 'auto_leds': False}
        d = {'tx0': (np.array([0, 1]), np.array([0.5, 0])), 'rx0_en': (np.array([2, 3]), np.array([1, 0]))}

        def change_lo(e):
            e.run() # compile internally
            e.set_lo_freq(2)
            return e.run() # compile internally

        # with warnings.catch_warnings():
        #     warnings.filterwarnings("ignore", category=RuntimeWarning) # catch GPA-FHDO error due to re-initialisation
        refl, siml = compare_expt_dict(d, "test_lo_change_expt", self.s, self.p, run_fn=change_lo, **expt_args)
        self.assertEqual(refl, siml)

if __name__ == "__main__":
    unittest.main()



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
from PyQt5.QtWidgets import QMainWindow, QTextEdit, QSizePolicy

class ConsoleWidget(QMainWindow):
    def __init__(self):
        super().__init__()

        self.console = QTextEdit()
        self.console.setReadOnly(True)
        self.setCentralWidget(self.console)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.setMaximumWidth(400)





"""
@author: J.M. Algarín, february 03th 2022
MRILAB @ I3M
"""


import time
import sys
import os
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

#*****************************************************************************
# Get the directory of the current script
main_directory = os.path.dirname(os.path.realpath(__file__))
parent_directory = os.path.dirname(main_directory)
parent_directory = os.path.dirname(parent_directory)

# Define the subdirectories you want to add to sys.path
subdirs = ['MaRGE', 'marcos_client']

# Add the subdirectories to sys.path
for subdir in subdirs:
    full_path = os.path.join(parent_directory, subdir)
    sys.path.append(full_path)
#******************************************************************************
import marge.controller.experiment_gui as ex
import numpy as np
from marge.seq.mriBlankSeq import MRIBLANKSEQ
import configs.hw_config as hw
import configs.units as units

class Noise(MRIBLANKSEQ):
    def __init__(self):
        super().__init__()

        # Input the parameters
        self.repetitionTime = None
        self.rxChannel = None
        self.nPoints = None
        self.bw = None
        self.freqOffset = None
        self.addParameter(key='seqName', string='NoiseInfo', val='Noise')
        self.addParameter(key='toMaRGE', val='True')
        self.addParameter(key='freqOffset', string='RF frequency offset (kHz)', val=0.0, units=units.kHz, field='RF')
        self.addParameter(key='nPoints', string='Number of points', val=2500, field='RF')
        self.addParameter(key='bw', string='Acquisition bandwidth (kHz)', val=50.0, units=units.kHz, field='RF')
        self.addParameter(key='rxChannel', string='Rx channel', val=0, field='RF')
        self.addParameter(key='repetitionTime', string='Repetition time (ms)', val=500.0, field='RF', units=units.ms)

    def sequenceInfo(self):
        
        print("Noise")
        print("Author: Dr. J.M. Algarín")
        print("Contact: josalggui@i3m.upv.es")
        print("mriLab @ i3M, CSIC, Spain")
        print("Get a noise measurement\n")
        

    def sequenceTime(self):
        return(0)  # minutes, scanTime

    def sequenceRun(self, plotSeq=0, demo=False):
        init_gpa = False
        self.demo = demo

        # Fix units to MHz and us
        self.freqOffset *= 1e-6  # MHz
        self.bw *= 1e-6  # MHz
        self.repetitionTime *= 1e6  # us

        self.mapVals['larmorFreq'] = hw.larmorFreq

        if self.demo:
            dataR = np.random.randn((self.nPoints + 2 * hw.addRdPoints) * hw.oversamplingFactor)
            dataC = np.random.randn((self.nPoints + 2 * hw.addRdPoints) * hw.oversamplingFactor)
            data = dataR+1j*dataC
            data = self.decimate(data_over=data, n_adc=1, option='Normal')
            acqTime = self.nPoints/self.bw
            tVector = np.linspace(0, acqTime, num=self.nPoints) * 1e-3  # ms
            spectrum = np.fft.ifftshift(np.fft.ifftn(np.fft.ifftshift(data)))
            fVector = np.linspace(-self.bw / 2, self.bw / 2, num=self.nPoints) * 1e3  # kHz
            self.dataTime = [tVector, data]
            self.dataSpec = [fVector, spectrum]
            time.sleep(self.repetitionTime*1e-6)
        else:
            samplingPeriod = 1 / self.bw
            self.expt = ex.Experiment(lo_freq=hw.larmorFreq + self.freqOffset,
                                      rx_t=samplingPeriod,
                                      init_gpa=init_gpa,
                                      gpa_fhdo_offset_time=(1 / 0.2 / 3.1),
                                      print_infos=False)
            samplingPeriod = self.expt.getSamplingRate()
            self.bw = 1/samplingPeriod
            acqTime = self.nPoints/self.bw

            # SEQUENCE
            self.iniSequence(20, np.array((0, 0, 0)))
            t0 = 30 + hw.addRdPoints*hw.oversamplingFactor/self.bw
            self.rxGateSync(t0, acqTime, channel=self.rxChannel)
            t0 = t0 + acqTime + hw.addRdPoints*hw.oversamplingFactor/self.bw
            if t0 < self.repetitionTime:
                self.endSequence(self.repetitionTime)
            else:
                self.endSequence(t0+20)

            # Load sequence to red pitaya
            if self.floDict2Exp():
                print("Sequence waveforms loaded successfully")
                pass
            else:
                print("ERROR: sequence waveforms out of hardware bounds")
                return False

            if plotSeq == 0:
                rxd, msgs = self.expt.run()
                data = self.decimate(rxd['rx%i' % self.rxChannel], 1, option='Normal')
                self.mapVals['data'] = data
                tVector = np.linspace(0, acqTime, num=self.nPoints) * 1e-3  # ms
                spectrum = np.fft.ifftshift(np.fft.ifftn(np.fft.ifftshift(data)))
                fVector = np.linspace(-self.bw / 2, self.bw / 2, num=self.nPoints) * 1e3  # kHz
                self.dataTime = [tVector, data]
                self.dataSpec = [fVector, spectrum]
            self.expt.__del__()

        return True

    def sequenceAnalysis(self, mode=None):
        self.mode = mode
        noiserms = np.std(self.dataTime[1])
        self.mapVals['RMS noise'] = noiserms
        self.mapVals['sampledPoint'] = noiserms # for sweep method
        bw = self.mapVals['bw'] * 1e3  # Hz
        noiserms = noiserms * 1e3  # uV
        print('rms noise: %0.1f uV @ %0.1f kHz' % (noiserms, bw * 1e-3))
        johnson = np.sqrt(2 * 50 * hw.temperature * bw * 1.38e-23) * 10 ** (hw.lnaGain / 20) * 1e6  # uV
        print('Expected by Johnson: %0.1f uV @ %0.1f kHz' % (johnson, bw * 1e-3))
        print('Noise factor: %0.1f johnson' % (noiserms / johnson))
        if noiserms / johnson > 4:
            print("WARNING: Noise is too high")

        # Plot signal versus time
        result1 = {'widget': 'curve',
                   'xData': self.dataTime[0],
                   'yData': [np.abs(self.dataTime[1]), np.real(self.dataTime[1]), np.imag(self.dataTime[1])],
                   'xLabel': 'Time (ms)',
                   'yLabel': 'Signal amplitude (mV)',
                   'title': 'Noise vs time',
                   'legend': ['abs', 'real', 'imag'],
                   'row': 0,
                   'col': 0}

        # Plot spectrum
        result2 = {'widget': 'curve',
                   'xData': self.dataSpec[0],
                   'yData': [np.abs(self.dataSpec[1])],
                   'xLabel': 'Frequency (kHz)',
                   'yLabel': 'Mag FFT (a.u.)',
                   'title': 'Noise spectrum',
                   'legend': [''],
                   'row': 1,
                   'col': 0}

        self.output = [result1, result2]

        self.saveRawData()

        if self.mode == 'Standalone':
            self.plotResults()

        return self.output


if __name__=='__main__':
    seq = Noise()
    seq.sequenceAtributes()
    seq.sequenceRun(demo=True)
    seq.sequenceAnalysis(mode='Standalone')




#!/usr/bin/env python3

import msgpack, warnings

from marcos.marcos_client.marmachine import MarServerWarning

version_major = 1
version_minor = 0
version_debug = 5
version_full = (version_major << 16) | (version_minor << 8) | version_debug

request_pkt = 0
emergency_stop_pkt = 1
close_server_pkt = 2
reply_pkt = 128

def construct_packet(data, packet_idx=0, command=request_pkt, version=(version_major, version_minor, version_debug)):
    vma, vmi, vd = version
    assert vma < 256 and vmi < 256 and vd < 256, "Version is too high for a byte!"
    version = (vma << 16) | (vmi << 8) | vd
    fields = [command, packet_idx, 0, version, data]
    return fields

# def process(payload, print_all=False):
#     # data = msgpack.unpackb(raw_reply, use_list=False, max_array_len=1024*1024)
#     reply_data = payload[4]

#     if print_all:
#         print("")

#         status = payload[5]

#         try:
#             print("Errors:")
#             for k in status['errors']:
#                 print(k)
#         except KeyError:
#             pass

#         try:
#             print("Warnings:")
#             for k in status['warnings']:
#                 print(k)
#         except KeyError:
#             pass

#         try:
#             print("Infos:")
#             for k in status['infos']:
#                 print(k)
#         except KeyError:
#             pass

#     try:
#         print("Last elements of returned unsigned arrays: {:f}, {:f}".format(
#             payload[4]['test_throughput']['array1'][-1], payload[4]['test_throughput']['array2'][-1]))
#     except KeyError:
#         print("Reply data: ")
#         print(reply_data)

def send_packet(packet, socket):
    socket.sendall(msgpack.packb(packet))

    unpacker = msgpack.Unpacker()
    packet_done = False
    while not packet_done:
        buf = socket.recv(1024)
        if not buf:
            break
        unpacker.feed(buf)
        for o in unpacker: # ugly way of doing it
            return o # quit function after 1st reply (could make this a thread in the future)

def command(server_dict, socket, print_infos=False, assert_errors=False):
    packet = construct_packet(server_dict)
    reply = send_packet(packet, socket)
    return_status = reply[5]

    if print_infos and 'infos' in return_status:
        print("Server info:")
        for k in return_status['infos']:
            print(k)

    if 'warnings' in return_status:
        for k in return_status['warnings']:
            warnings.warn(k, MarServerWarning)

    if 'errors' in return_status:
        if assert_errors:
            assert 'errors' not in return_status, return_status['errors'][0]
        else:
            for k in return_status['errors']:
                warnings.warn("SERVER ERROR: " + k, RuntimeWarning)

    return reply, return_status



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QLineEdit
from marge.widgets.widget_sequence_inputs import SequenceInputsWidget
from marge.seq.sequences import defaultsequences
import numpy as np


class SequenceInputsController(SequenceInputsWidget):
    """
    Controller class for managing sequence inputs.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Inherits:
        SequenceInputsWidget: Base class for sequence inputs widget.
    """
    def __init__(self, *args, **kwargs):
        """
        Initializes the SequenceInputsController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(SequenceInputsController, self).__init__(*args, **kwargs)
        self.seq_name = self.main.sequence_list.seq_name
        self.displayInputParameters()

    def displayInputParameters(self):
        """
        Displays the input parameters for the selected sequence.
        """
        self.seq_name = self.main.sequence_list.seq_name
        if hasattr(defaultsequences[self.seq_name], 'IMproperties'):
            props, tips = defaultsequences[self.seq_name].IMproperties
            input_widgets_1 = self.createTab(props, tips)
            self.addTab(input_widgets_1, "Image")

        if hasattr(defaultsequences[self.seq_name], 'RFproperties'):
            props, tips = defaultsequences[self.seq_name].RFproperties
            input_widgets_2 = self.createTab(props, tips)
            self.addTab(input_widgets_2, "RF")

        if hasattr(defaultsequences[self.seq_name], 'SEQproperties'):
            props, tips = defaultsequences[self.seq_name].SEQproperties
            input_widgets_3 = self.createTab(props, tips)
            self.addTab(input_widgets_3, "Sequence")

        if hasattr(defaultsequences[self.seq_name], 'OTHproperties'):
            props, tips = defaultsequences[self.seq_name].OTHproperties
            input_widgets_4 = self.createTab(props, tips)
            self.addTab(input_widgets_4, "Others")

        if hasattr(defaultsequences[self.seq_name], 'PROproperties'):
            props, tips = defaultsequences[self.seq_name].PROproperties
            input_widgets_5 = self.createTab(props, tips)
            self.addTab(input_widgets_5, "Processing")

    def removeTabs(self):
        """
        Removes all the tabs from the sequence inputs widget.
        """
        self.removeTab(0)
        self.removeTab(0)
        self.removeTab(0)
        self.removeTab(0)
        self.removeTab(0)

    def createTab(self, inputs, tips):
        """
        Creates a tab for the given inputs and tips.

        Args:
            inputs (dict): The input parameters.
            tips (dict): The tips for the input parameters.

        Returns:
            QWidget: The created tab widget.
        """
        # Create widget
        widget = QWidget()

        # Inputs are distributed in vertical layout
        layout = QVBoxLayout()
        widget.setLayout(layout)

        # Add inputs. Label and Values are distributed in horizontal layout
        for key in inputs.keys():
            input_layout = SequenceParameter([key, str(inputs[key][0])], tips[key][0], self.seq_name)
            layout.addLayout(input_layout)

        # Add spacer to place all the items on the top
        layout.addStretch()

        return widget


class SequenceParameter(QHBoxLayout):
    """
    Custom QHBoxLayout class for managing sequence parameters.

    Args:
        values (list): The parameter values.
        tip (str): The status tip for the input value.
        seq_name (str): The sequence name.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Inherits:
        QHBoxLayout: Base class for horizontal layout.
    """
    def __init__(self, values, tip, seq_name, *args, **kwargs):
        """
        Initializes the SequenceParameter.

        Args:
            values (list): The parameter values.
            tip (str): The status tip for the input value.
            seq_name (str): The sequence name.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(SequenceParameter, self).__init__(*args, **kwargs)
        self.seq_name = seq_name
        self.param_text = values[0]
        self.param_value = values[1]

        # QLabel
        self.input_label = QLabel(values[0])
        self.input_label.setMinimumWidth(150)
        self.addWidget(self.input_label)

        # QLineEdit
        self.input_value = QLineEdit(values[1])
        self.input_value.setStatusTip(tip)
        self.addWidget(self.input_value)
        self.input_value.textChanged.connect(self.getValue)

    def getValue(self):
        """
        Gets the input value and updates the sequence property mapVals.
        """
        sequence = defaultsequences[self.seq_name]

        # Get key for corresponding modified parameter
        names = sequence.mapNmspc  # Map with GUI names
        mod_name = self.input_label.text()  # GUI name of the modified value
        key = [k for k, v in names.items() if v == mod_name][0]  # Get corresponding key of the modified value
        valOld = sequence.mapVals[key]  # Current value to be saved again in case of error
        dataLen = sequence.mapLen[key]
        valNew = self.input_value.text()
        valNew = valNew.replace('[', '')
        valNew = valNew.replace(']', '')
        valNew = valNew.split(',')
        if type(valOld) == str:
            valOld = [valOld]
        elif dataLen == 1:
            valOld = [valOld]
        dataType = type(valOld[0])

        inputNum = []
        for ii in range(dataLen):
            if dataType == float or dataType == np.float64:
                try:
                    inputNum.append(float(valNew[ii]))
                except:
                    inputNum.append(float(valOld[ii]))
            elif dataType == int:
                try:
                    inputNum.append(int(valNew[ii]))
                except:
                    inputNum.append(int(valOld[ii]))
            else:
                try:
                    inputNum.append(str(valNew[0]))
                    break
                except:
                    inputNum.append(str(valOld[0]))
                    break
        if dataType == str:
            sequence.mapVals[key] = inputNum[0]
        else:
            if dataLen == 1:  # Save value into mapVals
                sequence.mapVals[key] = inputNum[0]
            else:
                sequence.mapVals[key] = inputNum

        # Print value into the console
        seqTime = sequence.sequenceTime()
        if hasattr(self, 'oldSeqTime'):
            if seqTime != self.oldSeqTime:
                print('Sequence time %0.1f minutes' % seqTime)
                self.oldSeqTime = seqTime
        else:
            print('Sequence time %0.1f minutes' % seqTime)
            self.oldSeqTime = seqTime

        defaultsequences[self.seq_name] = sequence



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import numpy as np
import copy

from marge.widgets.widget_smith_chart import PlotSmithChartWidget


class PlotSmithChartController(PlotSmithChartWidget):
    """
    TODO
    """
    def __init__(self,
                 x_data,  # numpy array
                 y_data,  # list of numpy array
                 legend,  # list of strings
                 x_label,  # string
                 y_label,  # string
                 title,  # string
                 ):
        """
        TODO
        """
        super(PlotSmithChartController, self).__init__()
        self.y_data = copy.copy(y_data)
        self.x_data = copy.copy(x_data)
        self.x_label = x_label
        self.y_label = y_label
        self.title = title

        # Set text
        self.label2.setText("<span style='font-size: 8pt'>%s=%0.2f, %s=%0.2f</span>" % (x_label, 0, y_label, 0))

        # Create here the smith chart:
        theta = np.linspace(0, 2*np.pi, 100)
        x0 = np.cos(theta)
        y0 = np.sin(theta)
        x1 = 0.5 + 0.5 * np.cos(theta)
        y1 = 0.5 * np.sin(theta)
        x2 = np.array([1, -1, 1])
        y2 = np.array([0, 0, 0])
        x3 = 0.1*np.cos(theta)
        y3 = 0.1*np.sin(theta)
        x_smith = np.concatenate((x0, x1, x2, x3), axis=0)
        y_smith = np.concatenate((y0, y1, y2, y3), axis=0)
        l_smith = 'Smith chart'

        # Add smith chart to plots
        if type(self.x_data) is list:
            self.x_data.append(x_smith)
        else:
            self.x_data = [self.x_data]
            self.x_data.append(x_smith)
        self.y_data.append(y_smith)
        legend.append(l_smith)

        # Add lines to plot_item
        n_lines = len(self.y_data)
        self.lines = []
        x_min = 0
        x_max = 0
        y_min = 0
        y_max = 0
        for line in range(n_lines):
            if type(self.x_data) is list:
                x = self.x_data[line]
            else:
                x = self.x_data.copy()
            y = self.y_data[line]
            if line == 0:
                self.lines.append(self.plot_item.plot(x, y, pen=self.pen[line], name=legend[line], symbol='o'))
                x_min = np.min(x)
                x_max = np.max(x)
            elif line == 2:
                self.lines.append(self.plot_item.plot(x, y, pen=self.pen[6], name=legend[line]))
            else:
                self.lines.append(self.plot_item.plot(x, y, pen=self.pen[line], name=legend[line]))
            if np.min(x) < x_min:
                x_min = np.min(x)
            if np.max(x) > x_max:
                x_max = np.max(x)
            if np.min(y) < y_min:
                y_min = np.min(y)
            if np.max(y) > y_max:
                y_max = np.max(y)
            self.plot_item.setXRange(x_min, x_max, padding=0)
            self.plot_item.setYRange(y_min, y_max, padding=0)
            if y_min == y_max:
                self.plot_item.setYRange(-1, 1, padding=0)

        # Set the plot properties
        self.plot_item.setTitle("%s" % title)
        self.plot_item.setLabel('bottom', x_label)
        self.plot_item.setLabel('left', y_label)

    def mouseMoved(self, evt):
        """
        Handle the mouseMoved event and update the information displayed at the cursor position on the plot.

        Args:
            evt (QGraphicsSceneMouseEvent): The mouse event object.

        Returns:
            None
        """
        pos = evt[0]
        if self.plot_item.sceneBoundingRect().contains(pos):
            if type(self.x_data) is not list:
                curves = self.plot_item.listDataItems()
                x, y = curves[0].getData()
                mouse_point = self.plot_item.vb.mapSceneToView(pos)
                index = np.argmin(np.abs(self.x_data - mouse_point.x()))
                self.label2.setText("<span style='font-size: 8pt'>%s=%0.2f, %s=%0.2f</span>" % (self.x_label,
                                                                                                x[index],
                                                                                                self.y_label,
                                                                                                y[index]))
                self.crosshair_v.setPos(x[index])
                self.crosshair_h.setPos(y[index])
            else:
                mouse_point = self.plot_item.vb.mapSceneToView(pos)
                self.label2.setText("x = %0.4f, y = %0.4f" % (mouse_point.x(), mouse_point.y()))
                self.crosshair_v.setPos(mouse_point.x())
                self.crosshair_h.setPos(mouse_point.y())



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
from PyQt5.QtWidgets import QToolBar, QAction
from PyQt5.QtGui import QIcon

class MarcosToolBar(QToolBar):
    def __init__(self, main, *args, **kwargs):
        super(MarcosToolBar, self).__init__(*args, **kwargs)
        self.main = main

        # Setup all
        self.action_start = QAction(QIcon("resources/icons/initGPA.png"), "Setup MaRCoS", self)
        self.action_start.setStatusTip("Setup MaRCoS")
        self.addAction(self.action_start)

        # Setup MaRCoS
        self.action_copybitstream = QAction(QIcon("resources/icons/M.png"), "MaRCoS init", self)
        self.action_copybitstream.setStatusTip("Install MaRCoS into Red Pitaya")
        self.addAction(self.action_copybitstream)

        # Connect to the server
        self.action_server = QAction(QIcon("resources/icons/server-light.png"), "MaRCoS server", self)
        self.action_server.setStatusTip("Connect to server")
        self.addAction(self.action_server)

        # Setup GPA board
        self.action_gpa_init = QAction(QIcon("resources/icons/gpa.png"), "Init power modules", self)
        self.action_gpa_init.setStatusTip("Init GPA board")
        self.addAction(self.action_gpa_init)
        


from setuptools import setup, find_packages

setup(
    name="marge-mri-example",
    version="0.0.5",
    packages=find_packages(),
    install_requires=[],
    author="Nihal Ouahabi",
    description="Une description de ton package MaRGE",
)




import datetime
import sys
import os
import atexit

from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot
from marge.widgets.widget_console import ConsoleWidget


class ConsoleController(ConsoleWidget):
    def __init__(self, log_name=None):
        super().__init__()

        # Créer le dossier logs si besoin
        log_folder = "logs"
        if not os.path.exists(log_folder):
            os.makedirs(log_folder)
            print(f"[ConsoleController] Created log folder at: {log_folder}")
        else:
            print(f"[ConsoleController] Log folder already exists at: {log_folder}")

        # Créer un nouveau nom de fichier log à chaque démarrage
        if log_name is None:
            pass
        else:
            self.log_filename = os.path.join(log_folder, log_name)

        # Ouvrir le fichier log
        try:
            self.log_file = open(self.log_filename, "a", encoding="utf-8")
            print(f"[ConsoleController] Logging to file: {self.log_filename}")
        except Exception as e:
            self.log_file = None
            # print(f"[ConsoleController] Failed to open log file: {e}")

        # Rediriger stdout
        sys.stdout = EmittingStream(textWritten=self.write_console)

        print("READY - MaRGE has started successfully.")
        print("WARNING - This is a test warning message.")
        print("ERROR - This is a test error message.")
        atexit.register(self.close_log)

    def write_console(self, text):
        # Console GUI
        cursor = self.console.textCursor()
        cursor.movePosition(cursor.End)

        current_time = datetime.datetime.now().strftime("%H:%M:%S")
        timestamp_html = f"<b>{current_time}</b>"

        if "ERROR" in text:
            text = f'<span style="color:red;"><b>ERROR</b></span>{text[5:]}'
        elif "WARNING" in text:
            text = f'<span style="color:orange;"><b>WARNING</b></span>{text[7:]}'
        elif "READY" in text:
            text = f'<span style="color:green;"><b>READY</b></span>{text[5:]}'

        if text.strip() == "":
            formatted_text = text
        else:
            formatted_text = f"{timestamp_html} - {text}".replace("\n", "<br>")

        if "<b>" in formatted_text and "</b>" in formatted_text:
            cursor.insertHtml(formatted_text)
        else:
            cursor.insertText(formatted_text)

        self.console.setTextCursor(cursor)
        self.console.ensureCursorVisible()

        # Nettoyage HTML pour le log
        clean_text = (
            text.replace("<br>", "\n")
                .replace("<b>", "")
                .replace("</b>", "")
                .replace('<span style="color:red;">', '')
                .replace('<span style="color:orange;">', '')
                .replace('<span style="color:green;">', '')
                .replace('</span>', '')
        )

        if hasattr(self, 'log_file') and self.log_file and not self.log_file.closed:
            try:
                timestamp = datetime.datetime.now().strftime('%H:%M:%S')
                self.log_file.write(f"{timestamp} - {clean_text.strip()}\n")
                self.log_file.flush()
            except Exception as e:
                print(f"[ConsoleController] Log write error: {e}")

    def clear_console(self):
        self.console.clear()

    def close_log(self):
        if hasattr(self, 'log_file') and self.log_file and not self.log_file.closed:
            self.log_file.close()


class EmittingStream(QObject):
    textWritten = pyqtSignal(str)

    def write(self, text):
        self.textWritten.emit(str(text))

    @pyqtSlot()
    def flush(self):
        pass



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
from marge.widgets.widget_sequence_list import SequenceListWidget
from marge.seq.sequences import defaultsequences


class SequenceListController(SequenceListWidget):
    """
    Controller class for managing the sequence list.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Inherits:
        SequenceListWidget: Base class for sequence list widget.
    """
    def __init__(self, *args, **kwargs):
        """
        Initializes the SequenceListController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(SequenceListController, self).__init__(*args, **kwargs)
        self.seq_name = self.getCurrentSequence()
        self.currentTextChanged.connect(self.updateSequence)
        self.currentTextChanged.connect(self.showSequenceInfo)

        # Here the GUI updates the inputs to the last used inputs
        for key in defaultsequences.keys():
            defaultsequences[key].loadParams()

    def getCurrentSequence(self):
        """
        Returns the name of the currently selected sequence.

        Returns:
            str: The name of the current sequence.
        """
        return self.currentText()

    def updateSequence(self):
        """
        Updates the selected sequence and resets the corresponding input parameters.
        """
        # Get the name of the selected sequence
        self.seq_name = self.getCurrentSequence()

        # Reset the tabs with corresponding input parameters
        if hasattr(self.main, "sequence_inputs"):
            self.main.sequence_inputs.removeTabs()
            self.main.sequence_inputs.displayInputParameters()

    def showSequenceInfo(self):
        """
        Displays information about the selected sequence.
        """
        # Get the name of the selected sequence
        self.seq_name = self.getCurrentSequence()

        defaultsequences[self.seq_name].sequenceInfo()



import bm4d
import threading
import numpy as np
from scipy.ndimage import gaussian_filter
from marge.widgets.widget_post import PostProcessingTabWidget
from skimage.util import view_as_blocks
from skimage.measure import shannon_entropy


class PostProcessingTabController(PostProcessingTabWidget):
    """
    Controller class for the post-processing tab widget.

    Inherits from PostProcessingTabWidget.

    Attributes:
        image_data (np.ndarray): The input image data.
        denoised_image (np.ndarray): The denoised image data.
        run_filter_button (QPushButton): QPushButton for applying BM4D filter.
        run_gaussian_button (QPushButton): QPushButton for applying Gaussian filter.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the PostProcessingTabController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(PostProcessingTabController, self).__init__(*args, **kwargs)

        # Connect the buttons to their corresponding methods
        self.run_filter_button.clicked.connect(self.bm4dFilter)
        self.run_gaussian_button.clicked.connect(self.gaussianFilter)

        self.image_data = None
        self.denoised_image = None

    def bm4dFilter(self):
        """
        Perform the BM4D filter operation using threading.

        Starts a new thread to execute the runBm4dFilter method.
        """

        thread = threading.Thread(target=self.runBm4dFilter)
        thread.start()

    def runBm4dFilter(self):
        """
        Run the BM4D filter operation.

        Retrieves the image data, performs rescaling, computes median and median absolute deviation (MAD),
        applies the BM4D filter to the rescaled image, restores the denoised image to its original dimensions,
        updates the main matrix of the image view widget, adds the operation to the history widget,
        and updates the operations history.
        """
        print('BM4D is loading')

        # Get the absolute value of the main image matrix and convert it to float
        image_data = np.abs(self.main.image_view_widget.main_matrix).astype(float)
        # Rescale the image data to the range (0, 100)
        reference = np.max(image_data)
        image_rescaled = image_data/reference*100

        # Calculate the standard deviation (sigma_psd) for BM4D filter
        if self.auto_checkbox.isChecked():
            # Quantize image
            num_bins = 1000
            image_quantized = np.digitize(image_rescaled, bins=np.linspace(0, 1, num_bins + 1)) - 1


            # Divide the image into blocks
            n_multi = (np.array(image_quantized.shape) / 5).astype(int) * 5
            blocks_q = view_as_blocks(image_quantized[0:n_multi[0], 0:n_multi[1], 0:n_multi[2]], block_shape=(5, 5, 5))
            blocks_r = view_as_blocks(image_rescaled[0:n_multi[0], 0:n_multi[1], 0:n_multi[2]], block_shape=(5, 5, 5))

            # Calculate the standard deviation for each block
            block_std_devs = np.std(blocks_r, axis=(3, 4, 5))

            # Calculate the average value for each block
            block_mean = np.mean(blocks_r, axis=(3, 4, 5))

            # Calculate entropy for each block
            block_entropies = np.zeros_like(blocks_q[:, :, :, 0, 0, 0], dtype=np.float32)
            for ii in range(blocks_q.shape[0]):
                for jj in range(blocks_q.shape[1]):
                    for kk in range(blocks_q.shape[2]):
                        block = blocks_q[ii, jj, kk, :, :, :]
                        entropy = shannon_entropy(block)
                        block_entropies[ii, jj, kk] = entropy

            # Find the indices of the block with the highest entropy
            max_entropy_index = np.unravel_index(np.argmax(block_entropies), block_entropies.shape)

            # Find the indices of the block with the minimum mean
            min_mean_index = np.unravel_index(np.argmin(block_mean), block_mean.shape)

            # Extract the block with the highest entropy from the block_std_devs array
            std = 3 * block_std_devs[min_mean_index]
            print("Standard deviation for BM4D: %0.2f" % std)

        else:
            std = float(self.std_text_field.text())

        # Create a BM4D profile and set the stage argument and blockmatches options
        profile = bm4d.BM4DProfile()
        stage_arg = bm4d.BM4DStages.ALL_STAGES
        blockmatches = (False, False)

        # Apply the BM4D filter to the rescaled image
        denoised_rescaled = bm4d.bm4d(image_rescaled, sigma_psd=std, profile=profile, stage_arg=stage_arg,
                                      blockmatches=blockmatches)

        # Rescale the denoised image back to its original dimensions
        denoised_image = denoised_rescaled/100*reference

        # Update the main image view widget with the denoised image
        self.main.image_view_widget.main_matrix = denoised_image

        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="BM4D",
                                          image=self.main.image_view_widget.main_matrix,
                                          orientation=self.main.toolbar_image.mat_data['axesOrientation'][0],
                                          operation="BM4D - Standard deviation: %0.2f" % std,
                                          space="i",
                                          image_key=self.main.image_view_widget.image_key)

        print('BM4D filter has been applied')

        # Update the space dictionary
        self.main.history_list.space[self.main.history_list.image_key] = 'i'

    def gaussianFilter(self):
        """
        Perform the Gaussian filter operation using threading.

        Starts a new thread to execute the runGaussianFilter method.
        """

        thread = threading.Thread(target=self.runGaussianFilter)
        thread.start()

    def runGaussianFilter(self):
        """
        Run the Gaussian filter operation.

        Retrieves the image data, applies the Gaussian filter, restores the filtered image to its original dimensions,
        updates the main matrix of the image view widget, adds the operation to the history widget,
        and updates the operations history.
        """
        # Get the absolute value of the main image matrix and convert it to float
        image_data = np.abs(self.main.image_view_widget.main_matrix).astype(float)

        # Get the standard deviation (sigma) for the Gaussian filter
        sigma = float(self.gaussian_text_field.text())

        # Rescale the image data to the range (0, 100)
        image_rescaled = np.interp(image_data, (np.min(image_data), np.max(image_data)), (0, 100))

        # Apply the Gaussian filter to the rescaled image
        filtered_image = gaussian_filter(image_data, sigma=sigma)

        # Rescale the filtered image back to its original dimensions
        gaussian_image = np.interp(filtered_image, (np.min(filtered_image), np.max(filtered_image)),
                                   (np.min(image_data), np.max(image_data)))

        # Update the main image view widget with the filtered image
        self.main.image_view_widget.main_matrix = gaussian_image

        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="Gaussian",
                                          image=self.main.image_view_widget.main_matrix,
                                          orientation=self.main.toolbar_image.mat_data['axesOrientation'][0],
                                          operation="Gaussian - Standard deviation: " + str(sigma),
                                          space="i",
                                          image_key=self.main.image_view_widget.image_key)

        # Update the space dictionary
        self.main.history_list.space[self.main.history_list.image_key] = 'i'



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import csv
import os
import platform
import shutil

from PyQt5.QtWidgets import QFileDialog

from marge.seq.sequences import defaultsequences
from marge.widgets.widget_toolbar_protocols import ProtocolsToolBar


class ProtocolsController(ProtocolsToolBar):
    """
    Controller class for managing protocols in the application.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Inherits:
        ProtocolsToolBar: Base class for the protocols toolbar.
    """
    def __init__(self, *args, **kwargs):
        """
        Initializes the ProtocolsController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)

        self.action_new_protocol.triggered.connect(self.newProtocol)
        self.action_del_protocol.triggered.connect(self.delProtocol)
        self.action_new_sequence.triggered.connect(self.newSequence)
        self.action_del_sequence.triggered.connect(self.delSequence)

    def delProtocol(self):
        """
        Deletes a protocol.

        Opens a file dialog to select a protocol directory for removal.
        """
        # Open a file dialog to get the filename to save to
        directory = 'protocols'
        folder_name = QFileDialog.getExistingDirectory(self.main, "Remove protocol", directory)

        if folder_name:
            shutil.rmtree(folder_name)
            print("Protocol removed")
            self.main.protocol_list.updateProtocolList()

    def delSequence(self):
        """
        Deletes a sequence from a protocol.

        Opens a file dialog to select a sequence file for removal from the current protocol.
        """
        # Get the current protocol
        protocol = self.main.protocol_list.getCurrentProtocol()

        # Open a file dialog to get the filename to save to
        directory = 'protocols/%s' % protocol
        file_name, _ = QFileDialog.getOpenFileName(self.main, 'Remove sequence from protocol', directory, '(*.csv)')

        # Delete protocol
        if file_name:
            os.remove(file_name)
            print("Protocol removed")
            self.main.protocol_inputs.updateProtocolInputs()

    def newProtocol(self):
        """
        Creates a new protocol.

        Opens a file dialog to specify the name and location for the new protocol.
        """
        # Open a file dialog to get the filename to save to
        file_name, _ = QFileDialog.getSaveFileName(self.main, 'New Protocol', 'protocols', '')

        if file_name:
            # Delete extension
            file_name = file_name.split('.')[0]

            # Check if the folder is the good one
            directory = os.path.dirname(file_name).split('/')[-1]
            protocol = file_name.split('/')[-1]
            if directory != 'protocols':
                print("Error. New protocols should be in 'protocols' folder.")
                return

            if not os.path.exists(file_name):
                os.makedirs(file_name)
                print("New protocol created successfully")
                self.main.protocol_inputs.sequences[protocol] = []
                self.main.protocol_list.updateProtocolList()
                self.main.protocol_inputs.updateProtocolInputs()
            else:
                print("Protocol already exist")

    def newSequence(self):
        """
        Adds a new sequence to a protocol.

        Gets the current protocol and sequence, opens a file dialog to specify the name and location
        for the new sequence file within the current protocol directory, and saves the sequence as a CSV file.
        """
        # Get the current protocol
        protocol = self.main.protocol_list.getCurrentProtocol()

        # Get the current sequence
        seq_name = self.main.sequence_list.getCurrentSequence()
        sequence = defaultsequences[seq_name]

        # Open a file dialog to get the filename to save to
        directory = 'protocols/%s' % protocol
        options = QFileDialog.Options()
        file_name, _ = QFileDialog.getSaveFileName(
            None, "Add sequence to protocol", directory, "",
            options=options
        )

        if file_name:
            if platform.system()=='Linux':
                file_name = "%s_%s.csv" % (seq_name, file_name.split('/')[-1])
            else:
                file_name = "%s_%s" % (seq_name, file_name.split('/')[-1])

            # Save csv with input parameters
            with open('%s/%s' % (directory, file_name), 'w') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=sequence.mapKeys)
                writer.writeheader()
                map_vals = {}
                for key in sequence.mapKeys:  # take only the inputs from mapVals
                    map_vals[key] = sequence.mapVals[key]
                writer.writerows([sequence.mapNmspc, map_vals])

            self.main.protocol_inputs.updateProtocolInputs()

            print("%s sequence added to the %s protocol" % (file_name.split('.')[0], protocol))
